from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from functools import partial
from typing import Tuple, cast, Optional
from decimal import Decimal as PyDecimal

from relationalai.early_access.metamodel.rewrite import Flatten
from relationalai.util.graph import topological_sort
from relationalai.early_access.metamodel import ir, compiler as c, visitor as v, factory as f, builtins, types, \
    helpers
from relationalai.early_access.metamodel.typer import Checker, InferTypes, typer
from relationalai.early_access.metamodel.builtins import from_cdc_annotation, concept_relation_annotation
from relationalai.early_access.metamodel.types import Hash, String, Number, Int, Decimal64, Decimal128, Bool, Date, DateTime, Float
from relationalai.early_access.metamodel.util import FrozenOrderedSet, OrderedSet, frozen, ordered_set, filter_by_type, \
    NameCache
from relationalai.early_access.devmode import sql, rewrite


class Compiler(c.Compiler):
    def __init__(self, skip_denormalization:bool=False):
        rewrites = [
            Checker(),
            InferTypes(),
            Flatten(handle_outputs=False, handle_union=False),
            rewrite.RecursiveUnion()
        ]
        if not skip_denormalization:
            # group updates, compute SCCs, use Sequence to denote their order
            rewrites.append(rewrite.Denormalize())
        super().__init__(rewrites)
        self.model_to_sql = ModelToSQL()

    def do_compile(self, model: ir.Model, options:dict={}) -> str:
        return str(self.model_to_sql.to_sql(model))


@dataclass
class ModelToSQL:
    """ Generates SQL from an IR Model, assuming the compiler rewrites were done. """

    relation_name_cache: NameCache = field(default_factory=NameCache)

    def to_sql(self, model: ir.Model) -> sql.Program:
        self._register_external_relations(model)
        return sql.Program(self._sort_dependencies(self._generate_statements(model)))

    def _generate_statements(self, model: ir.Model) -> list[sql.Node]:
        statements: list[sql.Node] = []
        for relation in model.relations:
            if self._is_table_creation_required(relation):
                statements.append(self._create_table(cast(ir.Relation, relation)))
        root = cast(ir.Logical, model.root)
        for child in root.body:
            if isinstance(child, ir.Logical):
                statements.extend(self._create_statement(cast(ir.Logical, child)))
            elif isinstance(child, ir.Union):
                statements.append(self._create_recursive_view(cast(ir.Union, child)))
        return statements

    #--------------------------------------------------
    # SQL Generation
    #--------------------------------------------------
    def _create_table(self, r: ir.Relation) -> sql.Node:
        return sql.CreateTable(
            sql.Table(self._relation_name(r),
                list(map(lambda f: sql.Column(f.name, self._convert_type(f.type)), r.fields))
            ))

    def _create_recursive_view(self, union: ir.Union) -> sql.Node:
        assert len(union.tasks) == 2
        assert isinstance(union.tasks[0], ir.Logical)
        assert isinstance(union.tasks[1], ir.Logical)

        def make_case_select(logical: ir.Logical):
            # TODO - improve the typing info to avoid these casts
            lookups = cast(list[ir.Lookup], filter_by_type(logical.body, ir.Lookup))
            # TODO - assuming a single update per case
            update = v.collect_by_type(ir.Update, logical).some()

            # TODO - rewriting references to the view, to use the CTE instead, with _rec
            new_lookups = []
            for lookup in lookups:
                if lookup.relation == update.relation:
                    new_lookups.append(f.lookup(
                        ir.Relation(f"{lookup.relation.name}_rec", lookup.relation.fields,
                                    frozen(), frozen()), lookup.args, lookup.engine))
                else:
                    new_lookups.append(lookup)

            aliases = []
            for i, arg in enumerate(update.args):
                aliases.append((update.relation.fields[i].name, arg))
            return self._make_select(new_lookups, aliases)

        # get a representative update
        update = v.collect_by_type(ir.Update, union).some()

        # TODO - maybe this should be more like INSERT INTO a table than a view?
        return sql.CreateView(self._relation_name(update.relation),
            sql.CTE(True, f"{self._relation_name(update.relation)}_rec", [f.name for f in update.relation.fields], [
                make_case_select(cast(ir.Logical, union.tasks[0])),
                make_case_select(cast(ir.Logical, union.tasks[1]))
            ]))

    def _create_statement(self, task: ir.Logical):

        # TODO - improve the typing info to avoid these casts
        nots = cast(list[ir.Not], filter_by_type(task.body, ir.Not))
        lookups = cast(list[ir.Lookup], filter_by_type(task.body, ir.Lookup))
        updates = cast(list[ir.Update], filter_by_type(task.body, ir.Update))
        outputs = cast(list[ir.Output], filter_by_type(task.body, ir.Output))
        logicals = cast(list[ir.Logical], filter_by_type(task.body, ir.Logical))
        constructs = cast(list[ir.Construct], filter_by_type(task.body, ir.Construct))
        ranks = cast(list[ir.Rank], filter_by_type(task.body, ir.Rank))
        aggs = cast(list[ir.Aggregate], filter_by_type(task.body, ir.Aggregate))
        unions = cast(list[ir.Union], filter_by_type(task.body, ir.Union))
        var_to_rank = {
            r.result: r
            for r in ranks
        } if ranks else {}
        var_to_agg = {
            (a.args[0] if a.aggregation == builtins.count else a.args[1]): a
            for a in aggs
        } if aggs else {}

        statements = []
        statements.extend(self._create_recursive_view(union) for union in unions)
        # TODO - this is simplifying soooo much :crying_blood:
        if updates and not lookups and not nots and not aggs:
            # TODO: this is assuming that the updates are all static values
            # insert static values: INSERT INTO ... VALUES ... or RawSource
            for u in updates:
                r = u.relation
                if r == builtins.raw_source:
                    lang, src = u.args[0], u.args[1]
                    if not (isinstance(lang, str) and lang.lower() == "sql"):
                        raise Exception(f"Unsupported language for RawSource: {lang}")
                    if not isinstance(src, str):
                        raise Exception(f"Expected SQL source to be a string, got: {type(src).__name__}")
                    statements.append(sql.RawSource(src))
                else:
                    tuples = self._get_tuples(task, u)
                    for tuple in tuples:
                        statements.append(
                            sql.Insert(self._relation_name(r), [f.name for f in r.fields], tuple, None)
                        )
        elif lookups or outputs or nots or aggs:
            # Some of the lookup relations we wrap into logical and we need to get them out for the SQL compilation.
            #    For example QB `decimal(0)` in IR will look like this:
            #        Logical ^[res]
            #           Exists(vDecimal128)
            #               Logical
            #                   int_to_decimal128(0, vDecimal128)
            #                   decimal128(vDecimal128, res)
            if logicals:
                lookups = [
                              lookup
                              for logical in logicals
                              for lookup in cast(list[ir.Lookup], filter_by_type(logical.body, ir.Lookup))
                          ] + lookups
            if updates:
                id_var_to_construct = {c.id_var: c for c in constructs} if constructs else {}
                # insert values that match a query: INSERT INTO ... SELECT ... FROM ... WHERE ...
                for u in updates:
                    r = u.relation
                    aliases = []
                    # We shouldn’t create or populate tables for value types that can be directly sourced from existing Snowflake tables.
                    if not self._is_value_type_population_relation(r):
                        for i, arg in enumerate(u.args):
                            field_name = r.fields[i].name
                            if isinstance(arg, ir.Var):
                                var_task = id_var_to_construct.get(arg) or var_to_rank.get(arg) or var_to_agg.get(arg)
                                if var_task:
                                    aliases.append((field_name, arg, var_task))
                                else:
                                    aliases.append((field_name, arg))
                            else:
                                aliases.append((field_name, arg))

                        statements.append(
                            sql.Insert(self._relation_name(r),
                                       [f.name for f in r.fields], [],
                                       self._make_select(lookups, aliases, nots, aggs, True)
                            )
                        )
            elif outputs:
                # output a query: SELECT ... FROM ... WHERE ...
                aliases = []
                distinct = False
                for output in outputs:
                    distinct = distinct or output.keys is None
                    for key, arg in output.aliases:
                        if isinstance(arg, ir.Var) and arg in var_to_rank:
                            aliases.append((key, arg, var_to_rank[arg]))
                        else:
                            aliases.append((key, arg))

                statements.append(self._make_select(lookups, aliases, nots, distinct=distinct))
        elif logicals:
            for logical in logicals:
                statements.extend(self._create_statement(logical))
        else:
            raise Exception(f"Cannot create SQL statement for:\n{task}")
        return statements

    def _make_select(self, lookups: list[ir.Lookup], outputs: list[Tuple[str, ir.Value]|Tuple[str, ir.Value, ir.Task]],
                     nots: Optional[list[ir.Not]] = None, aggs: Optional[list[ir.Aggregate]] = None, distinct: bool = False) -> sql.Select:

        table_lookups = OrderedSet.from_iterable(t for t in lookups if not builtins.is_builtin(t.relation))
        froms, wheres, sql_vars, var_column, var_sql_var, var_lookups = self._extract_lookups_metadata(table_lookups, 0)

        builtin_lookups = OrderedSet.from_iterable(t for t in lookups if builtins.is_builtin(t.relation))
        builtin_vars, builtin_wheres = self._resolve_builtins(builtin_lookups, var_lookups, var_column, var_sql_var)
        wheres.extend(builtin_wheres)

        wheres.extend(self._generate_where_clauses(var_lookups, var_column, sql_vars, builtin_vars))

        not_null_vars, vars = self._generate_select_output(outputs, builtin_vars, sql_vars, var_column,
                                                           var_lookups, var_sql_var)

        if not_null_vars:
            wheres.extend(sql.NotNull(var) for var in not_null_vars)

        not_exists, _ = self._generate_select_nots(nots, sql_vars, var_column, len(sql_vars))
        wheres.extend(not_exists)

        where = self._process_wheres_clauses(wheres)

        group_by = self._generate_aggs_group_by(aggs, builtin_vars, sql_vars, var_column, var_lookups)

        return sql.Select(distinct, vars, froms, where, group_by)

    def _extract_lookups_metadata(self, lookups: OrderedSet[ir.Lookup], start_index: int = 0):
        froms: list[sql.From] = []
        wheres: list[sql.Expr] = []
        sql_vars: dict[ir.Lookup, str] = dict()  # one var per table lookup
        var_column: dict[Tuple[ir.Var, ir.Lookup], ir.Field] = dict()
        var_sql_var: dict[ir.Var, str] = dict()
        var_lookups: dict[ir.Var, OrderedSet[ir.Lookup]] = defaultdict(OrderedSet)
        i = start_index

        for lookup in lookups:
            varname = f"v{i}"
            i += 1
            froms.append(sql.From(self._relation_name(lookup.relation), varname))
            sql_vars[lookup] = varname
            for j, arg in enumerate(lookup.args):
                if isinstance(arg, ir.Var):
                    var_column[arg, lookup] = lookup.relation.fields[j]
                    var_sql_var[arg] = varname
                    var_lookups[arg].add(lookup)
                # case when Literal is used as a relation argument: `test(1, x)`
                elif isinstance(arg, (int, str, float, bool, ir.Literal)):
                    ref = f"{sql_vars[lookup]}.{lookup.relation.fields[j].name}"
                    wheres.append(sql.Terminal(f"{ref} = {self._convert_value(arg, False)}"))

        return froms, wheres, sql_vars, var_column, var_sql_var, var_lookups

    def _var_reference(self, var_lookups: dict[ir.Var, OrderedSet[ir.Lookup]], var_sql_var: dict[ir.Var, str],
                       var_column: dict[Tuple[ir.Var, ir.Lookup], ir.Field], v) -> str:
        if isinstance(v, ir.Var):
            # TODO - assuming the built-in reference was grounded elsewhere
            lookup = var_lookups[v].some()
            return f"{var_sql_var[v]}.{var_column[(v, lookup)].name}"
        elif isinstance(v, ir.Literal):
            return self._convert_value(v.value)
        return f"'{v}'" if isinstance(v, str) else str(v)

    def _resolve_builtin_var(self, builtin_vars: dict[ir.Var, ir.Value], var):
        # We need recursive lookup because it maybe a case when we need to join more than 2 lookups.
        #    For example QB `a != decimal(0)` in IR will look like this:
        #        Logical ^[res]
        #           Exists(vDecimal128)
        #               Logical
        #                   int_to_decimal128(0, vDecimal128)
        #                   decimal128(vDecimal128, res)
        #        a != res
        #    But we need to convert it to `a != 0` in SQL.
        if isinstance(var, ir.Var) and var in builtin_vars:
            val = builtin_vars[var]
            return self._resolve_builtin_var(builtin_vars, val) if isinstance(val, ir.Var) else val
        return var

    def _resolve_builtins(self, builtin_lookups: OrderedSet[ir.Lookup], var_lookups: dict[ir.Var, OrderedSet[ir.Lookup]],
                          var_column: dict[Tuple[ir.Var, ir.Lookup], ir.Field], var_sql_var: dict[ir.Var, str]):
        wheres: list[sql.Expr] = []
        builtin_vars: dict[ir.Var, ir.Value] = {}

        intermediate_builtin_vars: set[ir.Var] = {
            arg for lookup in builtin_lookups
            for arg in lookup.args
            if isinstance(arg, ir.Var) and arg not in var_lookups
        }

        reference = partial(self._var_reference, var_lookups, var_sql_var, var_column)
        resolve_builtin_var = partial(self._resolve_builtin_var, builtin_vars)

        for lookup in builtin_lookups:
            args = lookup.args
            relation = lookup.relation
            relation_name = self._relation_name(relation)

            if relation == builtins.substring:
                assert len(args) == 4, f"Expected 4 args for `strings.substring`, got {len(args)}: {args}"

                # Unpack and process arguments
                lhs_raw, from_idx_raw, to_idx_raw, output = args
                assert isinstance(output, ir.Var), "Fourth argument (output) must be a variable"
                from_idx = self._convert_value(from_idx_raw)
                to_idx = self._convert_value(to_idx_raw)

                # Resolve the left-hand side expression
                lhs = lhs_raw if lhs_raw in intermediate_builtin_vars else reference(lhs_raw)
                left = self._var_to_expr(lhs, reference, resolve_builtin_var) if isinstance(lhs_raw, ir.Var) else str(lhs)

                # Calculate substring length: SQL is 1-based and end-inclusive
                substring_len = int(to_idx) - int(from_idx) + 2
                assert substring_len >= 0, f"Invalid substring range: from {from_idx} to {to_idx}"

                expr = f"substring({left}, {from_idx}, {substring_len})"
                builtin_vars[output] = expr
            else:
                # Assuming infix binary or ternary operators here
                lhs_raw, rhs_raw = args[0], args[1]
                lhs = lhs_raw if lhs_raw in intermediate_builtin_vars else reference(lhs_raw)
                rhs = rhs_raw if rhs_raw in intermediate_builtin_vars else reference(rhs_raw)
                if relation in builtins.string_binary_builtins:
                    left = self._var_to_expr(lhs_raw, reference, resolve_builtin_var)
                    if relation == builtins.num_chars and isinstance(rhs_raw, ir.Var):
                        builtin_vars[rhs_raw] = f"length({left})"
                    else:
                        right = self._var_to_expr(rhs_raw, reference, resolve_builtin_var)
                        if relation == builtins.starts_with:
                            expr = f"concat({right}, '%')" if isinstance(rhs_raw, ir.Var) else f"'{right}%'"
                        elif relation == builtins.ends_with:
                            expr = f"concat('%', {right})" if isinstance(rhs_raw, ir.Var) else f"'%{right}'"
                        elif relation == builtins.like_match:
                            expr = right if isinstance(rhs_raw, ir.Var) else f"'{right}'"
                        elif relation == builtins.contains:
                            expr = f"concat('%', {right}, '%')" if isinstance(rhs_raw, ir.Var) else f"'%{right}%'"
                        else:
                            raise Exception(f"Unsupported string builtin relation: {relation}")
                        wheres.append(sql.Like(left, expr))
                elif relation in builtins.conversion_builtins and isinstance(rhs_raw, ir.Var):
                    # For number conversion relations like `decimal_to_float(x, x_float)`
                    # we need to store mapping to the original value to map it back in the next builtin relation.
                    # example: a = 0.0 in the IR is (decimal_to_float(a, a_float)) and (a_float = 0.0),
                    #   but we will make it back `a = 0.0` in the SQL query.
                    builtin_vars[rhs_raw] = lhs
                elif helpers.is_type_box(lookup) and isinstance(rhs_raw, ir.Var):
                    # For type box relations we keep the raw var, and we will ground it later.
                    builtin_vars[rhs_raw] = lhs_raw
                else:
                    left = self._var_to_expr(lhs, reference, resolve_builtin_var)
                    right = self._var_to_expr(rhs, reference, resolve_builtin_var)

                    if len(args) == 3:
                        out_var = args[2]
                        if isinstance(out_var, ir.Var):
                            if relation != builtins.concat:
                                expr = f"{left} {relation_name} {right}"

                                # For example, when this is an intermediate result
                                # example: c = a - b in the IR is (a - b = d) and (d = c)
                                builtin_vars[out_var] = expr
                            else:
                                builtin_vars[out_var] = f"{relation_name}({left}, {right})"
                        else:
                            raise Exception(
                                f"Expected `ir.Var` type for the relation `{relation}` output but got `{type(out_var).__name__}`: {out_var}"
                            )
                    else:
                        # Replace intermediate vars with disjoined expressions
                        expr = f"{left} {relation_name} {right}"
                        wheres.append(sql.Terminal(expr))

        return builtin_vars, wheres

    def _generate_where_clauses(self, var_lookups: dict[ir.Var, OrderedSet[ir.Lookup]],
                                var_column: dict[Tuple[ir.Var, ir.Lookup], ir.Field], sql_vars: dict[ir.Lookup, str],
                                builtin_vars: dict[ir.Var, ir.Value]):
        wheres: list[sql.Expr] = []
        for arg, lookup_set in var_lookups.items():
            # if there are 2 lookups for the same variable, we need a join
            if len(lookup_set) > 1:
                refs = [f"{sql_vars[lu]}.{var_column[cast(ir.Var, arg), lu].name}" for lu in lookup_set]
                # join variable references pairwise (e.g. "x.id = y.id AND y.id = z.id")
                for lhs, rhs in zip(refs, refs[1:]):
                    wheres.append(sql.Terminal(f"{lhs} = {rhs}"))
            # case when we have a builtin operation as a relation argument
            #   example: `test(a - 1, b)` and we are handling here `a - 1` arg.
            elif len(lookup_set) == 1 and arg in builtin_vars:
                lookup = lookup_set[0]
                ref = f"{sql_vars[lookup]}.{var_column[cast(ir.Var, arg), lookup].name}"
                rhs_ref = self._resolve_builtin_var(builtin_vars, arg)
                if isinstance(rhs_ref, ir.Var) and rhs_ref in var_lookups:
                    # Case: result of type-boxing variable
                    rhs_lookup = var_lookups[rhs_ref].some()
                    rhs = f"{sql_vars[rhs_lookup]}.{var_column[rhs_ref, rhs_lookup].name}"
                else:
                    rhs = rhs_ref.name if isinstance(rhs_ref, ir.Var) else str(rhs_ref)
                wheres.append(sql.Terminal(f"{ref} = {rhs}"))
        return wheres

    def _process_wheres_clauses(self, wheres: list[sql.Expr]) -> Optional[sql.Where]:
        # conjunction of not_wheres
        if len(wheres) == 0:
            where = None
        elif len(wheres) == 1:
            where = sql.Where(wheres[0])
        else:
            where = sql.Where(sql.And(wheres))
        return where

    def _generate_aggs_group_by(self, aggs: Optional[list[ir.Aggregate]], builtin_vars: dict[ir.Var, ir.Value],
                                sql_vars: dict[ir.Lookup, str], var_column: dict[Tuple[ir.Var, ir.Lookup], ir.Field],
                                var_lookups: dict[ir.Var, OrderedSet[ir.Lookup]]) -> list[sql.VarRef]:
        group_by: list[sql.VarRef] = []
        if aggs:
            # After flatten it can be only one aggregation per rule.
            agg = aggs[0]
            for var in agg.group:
                resolved_var = var
                # Resolve the variable if it's the result of a builtin
                if var in builtin_vars:
                    # Case: result of type-boxing variable
                    resolved_var = self._resolve_builtin_var(builtin_vars, var)

                if isinstance(resolved_var, ir.Var) and resolved_var in var_lookups:
                    lookup = var_lookups[resolved_var].some()
                    group_by.append(sql.VarRef(sql_vars[lookup], var_column[resolved_var, lookup].name, None))
                else:
                    # We may group by a result of a builtin operation.
                    # Example: `sum(Person.age).per(strings.len(Person.name))`
                    var_ref = resolved_var.name if isinstance(resolved_var, ir.Var) else str(resolved_var)
                    group_by.append(sql.VarRef(var_ref, None, None))
        return group_by

    def _generate_select_output(self, outputs: list[Tuple[str, ir.Value]|Tuple[str, ir.Value, ir.Task]],
                                builtin_vars: dict[ir.Var, ir.Value], sql_vars: dict[ir.Lookup, str],
                                var_column: dict[Tuple[ir.Var, ir.Lookup], ir.Field],
                                var_lookups: dict[ir.Var, OrderedSet[ir.Lookup]], var_sql_var: dict[ir.Var, str]):
        reference = partial(self._var_reference, var_lookups, var_sql_var, var_column)
        resolve_builtin_var = partial(self._resolve_builtin_var, builtin_vars)
        def handle_lookup_var(var):
            lookup = var_lookups[var].some()
            vars.append(sql.VarRef(sql_vars[lookup], var_column[var, lookup].name, alias))
            if from_cdc_annotation in lookup.relation.annotations:
                not_null_vars.add(f"{sql_vars[lookup]}.{var_column[var, lookup].name}")
        # finally, compute what the select will return
        vars = []
        not_null_vars = ordered_set()
        for output in outputs:
            alias, var = output[0], output[1]
            task = output[2] if len(output) > 2 else None
            if isinstance(var, ir.Var):
                if var in var_lookups:
                    handle_lookup_var(var)
                elif var in builtin_vars:
                    var_ref = resolve_builtin_var(var)
                    if var_ref in var_lookups:
                        # Case: result of type-boxing variable
                        handle_lookup_var(var_ref)
                    else:
                        # Example: We may have `decimal(0)` in QB which turns in IR into:
                        #   (int_to_decimal128(0, vDecimal128) and decimal128(vDecimal128, res_3))
                        #   and we need to make it `0` in SQL.
                        var_ref = var_ref.name if isinstance(var_ref, ir.Var) else str(var_ref)
                        vars.append(sql.VarRef(var_ref, None, alias))
                elif task:
                    if isinstance(task, ir.Construct):
                        # Generate constructions like hash(`x`, `y`, TABLE_ALIAS.COLUMN_NAME) as `alias`
                        elements = []
                        for v in task.values:
                            if isinstance(v, ir.Var):
                                lookup = var_lookups[v].some()
                                lookup_var = f"{sql_vars[lookup]}.{var_column[v, lookup].name}"
                                elements.append(lookup_var)
                                if from_cdc_annotation in lookup.relation.annotations:
                                    not_null_vars.add(lookup_var)
                            else:
                                elements.append(self._convert_value(v, True))
                        vars.append(sql.VarRef(f"hash({', '.join(elements)})", None, alias))
                    elif isinstance(task, ir.Rank):
                        order_by_vars = []
                        for arg, is_ascending in zip(task.args, task.arg_is_ascending):
                            order_by_vars.append(sql.OrderByVar(reference(arg), is_ascending))
                        partition_by_vars = [reference(arg) for arg in task.group] if task.group else []
                        vars.append(sql.RowNumberVar(order_by_vars, partition_by_vars, alias))
                    elif isinstance(task, ir.Aggregate):
                        result_arg = task.projection[0] if task.aggregation == builtins.count else task.args[0]
                        result_arg = resolve_builtin_var(result_arg)
                        ref = reference(result_arg) if isinstance(result_arg, ir.Var) else str(result_arg)
                        vars.append(sql.VarRef(f"{task.aggregation.name}({ref})", None, alias))
            else:
                # TODO - abusing even more here, because var is a value!
                vars.append(sql.VarRef(str(var), None, alias))
        return not_null_vars, vars

    def _generate_select_nots(self, nots: Optional[list[ir.Not]], sql_vars: dict[ir.Lookup, str],
                              var_column:dict[Tuple[ir.Var, ir.Lookup], ir.Field], index: int) -> tuple[list[sql.NotExists], int]:
        not_exists = []
        if nots:
            for not_expr in nots:
                logical = cast(ir.Logical, not_expr.task)
                all_lookups = cast(list[ir.Lookup], filter_by_type(logical.body, ir.Lookup))
                logicals = cast(list[ir.Logical], filter_by_type(logical.body, ir.Logical))
                inner_nots = cast(list[ir.Not], filter_by_type(logical.body, ir.Not))

                # Some of the lookup relations we wrap into logical and we need to get them out for the SQL compilation.
                #    For example QB `decimal(0)` in IR will look like this:
                #        Logical ^[res]
                #           Exists(vDecimal128)
                #               Logical
                #                   int_to_decimal128(0, vDecimal128)
                #                   decimal128(vDecimal128, res)
                if logicals:
                    all_lookups = [
                                  lookup
                                  for logical in logicals
                                  for lookup in cast(list[ir.Lookup], filter_by_type(logical.body, ir.Lookup))
                              ] + all_lookups

                lookups = OrderedSet.from_iterable(t for t in all_lookups if not builtins.is_builtin(t.relation))
                froms, wheres, not_sql_vars, not_var_column, not_var_sql_var, not_var_lookups \
                    = self._extract_lookups_metadata(lookups, index)
                index += len(not_sql_vars)

                builtin_lookups = OrderedSet.from_iterable(t for t in all_lookups if builtins.is_builtin(t.relation))
                builtin_vars, builtin_wheres = self._resolve_builtins(builtin_lookups, not_var_lookups, not_var_column,
                                                                      not_var_sql_var)
                wheres.extend(builtin_wheres)

                # We need to join the not exists select with the outside select query context
                for arg, lookup_set in not_var_lookups.items():
                    if len(lookup_set) > 0:
                        lu = lookup_set[0]
                        lhs = f"{not_sql_vars[lu]}.{not_var_column[cast(ir.Var, arg), lu].name}"

                        # lookup the same var from the outside context to make the join
                        matching_lookup = next(
                            (lookup for (var, lookup) in var_column if var == arg),
                            None
                        )

                        if matching_lookup is not None:
                            rhs = f"{sql_vars[matching_lookup]}.{var_column[(arg, matching_lookup)].name}"
                            wheres.append(sql.Terminal(f"{lhs} = {rhs}"))

                wheres.extend(self._generate_where_clauses(not_var_lookups, not_var_column, not_sql_vars, builtin_vars))

                inner_not_exists, index = self._generate_select_nots(inner_nots, not_sql_vars, not_var_column, index)
                wheres.extend(inner_not_exists)

                where = self._process_wheres_clauses(wheres)
                not_exists.append(sql.NotExists(sql.Select(False, [1], froms, where)))

        return not_exists, index

    def _var_to_expr(self, var, reference, resolve_builtin_var):
        """
        Convert a variable to an expression string.
        """
        resolved = resolve_builtin_var(var)
        return reference(resolved) if isinstance(resolved, ir.Var) else str(resolved)

    def _get_tuples(self, logical: ir.Logical, u: ir.Update):
        """
        Get a list of tuples to perform this update.

        This function traverses the update args, assuming they contain only static values or
        variables bound to a construct task, and generates a list of tuples to insert. There
        may be multiple tuples because arguments can be lists of values bound to a field
        whose role is multi.
        """
        # TODO - this only works if the variable is bound to a Construct task, we need a more general approach.
        values = []
        for a in u.args:
            if isinstance(a, ir.Var):
                for t in logical.body:
                    if isinstance(t, ir.Construct) and t.id_var == a:
                        values.append(f"hash({', '.join([self._convert_value(v, True) for v in t.values])})")
                        break
            elif isinstance(a, FrozenOrderedSet):
                values.append(frozen(*[self._convert_value(v) for v in a]))
            else:
                values.append(self._convert_value(a))
        return self._product(values)

    def _product(self, values):
        """ Compute a cartesian product of values when the value is a FrozenOrderedSet. """
        # TODO - some pass needs to check that this is correct, i.e. that we are using a
        # FrozenOrderedSet only if the field is of role multi.
        tuples = [[]]
        for value in values:
            if isinstance(value, FrozenOrderedSet):
                tuples = [prev + [element] for prev in tuples for element in value]
            else:
                tuples = [prev + [value] for prev in tuples]
        return tuples

    def _convert_value(self, v, quote_numbers:bool=False) -> str:
        """ Convert the literal value in v to a SQL value."""
        if isinstance(v, str):
            return f"'{v}'"
        if isinstance(v, PyDecimal):
            return str(v)
        if isinstance(v, ir.ScalarType):
            return f"'{v.name}'"
        if isinstance(v, ir.Literal):
            return self._convert_value(v.value, quote_numbers)
        return v if not quote_numbers else f"'{v}'"

    BUILTIN_CONVERSION = {
        Hash: "DECIMAL(38, 0)",
        String: "TEXT",
        Number: "DOUBLE",
        Int: "INT",
        Decimal64: "DECIMAL(19, 6)",
        Decimal128: "DECIMAL(38, 10)",
        Bool: "BOOLEAN",
        Date: "DATE",
        DateTime: "DATETIME",
        Float: "FLOAT(53)",
    }
    def _convert_type(self, t: ir.Type) -> str:
        """ Convert the type t into the equivalent SQL type."""
        # entities become DECIMAL(38, 0)
        if not types.is_builtin(t) and not types.is_value_type(t):
            return "DECIMAL(38, 0)"

        # convert known builtins
        base_type = typer.to_base_primitive(t)
        if isinstance(base_type, ir.ScalarType) and base_type in self.BUILTIN_CONVERSION:
            return self.BUILTIN_CONVERSION[base_type]
        raise Exception(f"Unknown built-in type: {t}")

    def _is_table_creation_required(self, r: ir.Relation) -> bool:
        """ Check if the relation should be created as a table in SQL. """
        # Skip built-in relations, builtin overloads and CDC relations
        if (builtins.is_builtin(r) or from_cdc_annotation in r.annotations or builtins.is_annotation(r) or
                helpers.is_type_box(r)):
            return False
        # Skip value type population relations
        return not self._is_value_type_population_relation(r)

    @staticmethod
    def _is_value_type_population_relation(r: ir.Relation) -> bool:
        """ Check if the relation is a value type relation. """
        if not r.fields or len(r.fields) != 1:
            return False
        return types.is_value_type(r.fields[0].type) and concept_relation_annotation in r.annotations

    def _relation_name(self, relation: ir.Relation):
        if helpers.is_external(relation) or helpers.builtins.is_builtin(relation):
            return relation.name
        return self.relation_name_cache.get_name(relation.id, relation.name, helpers.relation_name_prefix(relation))

    def _register_external_relations(self, model: ir.Model):
        # force all external relations to get a name in the cache, so that internal relations
        # cannot use those names in _relation_name
        for r in model.relations:
            if helpers.is_external(r):
                self.relation_name_cache.get_name(r.id, r.name)

    # TODO: need to check if inserts may depend on any view and if so then sort them together.
    def _sort_dependencies(self, statements: list[sql.Node]) -> list[sql.Node]:
        """
            Sorts SQL statements to ensure proper execution order:
            1. CREATE TABLE statements
            2. INSERT statements (topologically sorted by dependencies)
            3. UPDATE statements
            3. Other statements except SELECT queries (e.g., CREATE VIEW, etc.)
            4. SELECT queries
        """
        create_tables = []
        inserts: dict[str, list[sql.Insert]] = defaultdict(list)
        updates = []
        miscellaneous_statements = []
        selects = []

        for statement in statements:
            if isinstance(statement, sql.CreateTable):
                create_tables.append(statement)
            elif isinstance(statement, sql.Insert):
                inserts[statement.table].append(statement)
            elif isinstance(statement, sql.Update):
                updates.append(statement)
            elif isinstance(statement, sql.Select):
                selects.append(statement)
            else:
                miscellaneous_statements.append(statement)

        sorted_inserts = self._sort_inserts_dependency_graph(inserts)

        return create_tables + sorted_inserts + updates + miscellaneous_statements + selects

    @staticmethod
    def _sort_inserts_dependency_graph(insert_statements: dict[str, list[sql.Insert]]) -> list[sql.Insert]:
        """ Topologic sort INSERT statements based on dependencies in their SELECT FROM clauses. """
        nodes = list(insert_statements.keys())
        edges = []

        for target_table, inserts in insert_statements.items():
            for insert in inserts:
                select = insert.select
                if select and select.froms:
                    for from_clause in select.froms:
                        edges.append((from_clause.table, target_table))

        sorted_tables = topological_sort(nodes, edges)

        sorted_inserts = []
        for table in sorted_tables:
            if table in insert_statements:
                sorted_inserts.extend(insert_statements[table])

        return sorted_inserts



import re

from relationalai.early_access.dsl.constants import VERBAL_PART_CONNECTION
from relationalai.early_access.dsl.core.utils import camel_to_snake, to_rai_way_string


# Re-adjusted code from dsl/utils.py - that version was used in the dsl
def build_relation_name_from_reading(reading):
    roles = re.findall(r'\{([^}]+)}', reading)
    text_without_concepts = re.sub(r'\{[^}]+}', '|', reading)
    verbal_parts = [text.strip() for text in text_without_concepts.split('|') if text.strip()]

    # Handle postfix on the first player
    if verbal_parts[0].startswith("-"):
    #     postfix = verbal_parts[0].split(" ")[0]
        verbal_parts[0] = verbal_parts[0].replace("-", "")
    # Handle prefix on the first player
    elif "-" in verbal_parts[0].split(" ")[0]:
        verbal_parts.remove(verbal_parts[0])

    rel_name = ""
    if len(roles) == 1:
        if len(verbal_parts) > 0:
            rel_name = to_rai_way_string(verbal_parts[0])
            rel_name = rel_name.replace("_is_", "_")
        if not rel_name:
            rel_name = camel_to_snake(roles[0])
    elif len(roles) == 2:
        s_role = roles[1]
        if len(verbal_parts) > 0:
            rel_name = to_rai_way_string(verbal_parts[0])
            if (verbal_parts[0].strip().startswith(("has", "is")) and not rel_name) or rel_name.endswith("_"):
                    rel_name += camel_to_snake(s_role)
            elif rel_name.endswith("has"):
                rel_name = rel_name.replace("has", camel_to_snake(s_role))
            elif rel_name.endswith("is"):
                rel_name = rel_name.replace("is", camel_to_snake(s_role))
        if verbal_parts[len(verbal_parts)-1].startswith("-"):
            rel_name += to_rai_way_string(verbal_parts[len(verbal_parts)-1])
    else:
        join_parts = [to_rai_way_string(verbal_parts[0], False)]
        for i in range(1, len(roles)):
            role = roles[i]
            if not join_parts[len(join_parts) - 1].endswith(camel_to_snake(role)):
                join_parts.append(camel_to_snake(role.split(":")[0])) # take a role name instead of type if any
            if i < len(roles) - 1 or i < len(verbal_parts):
                if verbal_parts[i].startswith("-"):
                    join_parts.append(to_rai_way_string(verbal_parts[i][1:], False))
                elif verbal_parts[i].endswith("-"):
                    join_parts.append(to_rai_way_string(verbal_parts[i][:-1], False))
                else:
                    join_parts.append(to_rai_way_string(verbal_parts[i], False))
        rel_name = VERBAL_PART_CONNECTION.join(join_parts)
    return rel_name



from __future__ import annotations

from collections import OrderedDict
from typing import Optional, Any, Sequence, Union, TypeVar

import relationalai.early_access.builder as qb
from relationalai import Config
from relationalai.early_access.dsl.bindings.common import Binding
from relationalai.early_access.dsl.bindings.csv import CsvTable
from relationalai.early_access.dsl.bindings.snowflake import SnowflakeTable
from relationalai.early_access.dsl.codegen.weaver import Weaver
from relationalai.early_access.dsl.core.utils import generate_stable_uuid, to_pascal_case
from relationalai.early_access.dsl.orm.constraints import Unique, Mandatory, RoleValueConstraint, Range, \
    ExclusiveSubtypeConstraint, InclusiveSubtypeConstraint, InclusiveRoleConstraint, ExclusiveRoleConstraint, \
    RingConstraint
from relationalai.early_access.dsl.orm.relationships import Relationship, Role
from relationalai.early_access.dsl.orm.types import Concept
from relationalai.early_access.dsl.snow.api import Executor
from relationalai.early_access.metamodel.util import OrderedSet

T = TypeVar('T', int, float, str)

class Model:
    def __init__(
            self,
            name: str,
            is_primary: bool = True,
            dry_run: bool = False,
            use_lqp: bool = False,
            use_sql: bool = False,
            strict: bool = False,
            wide_outputs: bool = False,
            config: Optional[Config] = None
    ):
        self.name = name
        self._qb_model = qb.Model(name, dry_run=dry_run, strict=strict, config=config, use_lqp=use_lqp, use_sql=use_sql, wide_outputs=wide_outputs)
        self.is_primary = is_primary
        self._constraints = OrderedSet()
        self._entity_to_id_relationship = OrderedDict()
        self._bindable_tables = OrderedDict()
        self._bindings = []
        self._executor = None
        self._weaver = None
        self._queries = OrderedSet()
        self.Enum = self._qb_model.Enum

    def guid(self):
        return generate_stable_uuid(self.name)

    def Concept(self, name: str, extends: list[Any] = [], identify_by:dict[str, Any]={}) -> Concept:
        name = to_pascal_case(name)
        return Concept(self, name, extends, identify_by)

    def qb_model(self):
        return self._qb_model

    def Relationship(self, reading: Any, short_name:str="") -> qb.Relationship:
        """
        Create a new relationship with the given reading.
        """
        return Relationship(self, reading, short_name=short_name)

    def constraint(self, constraint):
        # Generic constraint addition with validation
        self._constraints.add(constraint)
        constraint._desugar()

    def unique(self, *roles):
        self.constraint(Unique(*roles))

    def mandatory(self, role):
        self.constraint(Mandatory(role))

    def inclusive_roles(self, *roles):
        self.constraint(InclusiveRoleConstraint(*roles))

    def exclusive_roles(self, *roles):
        self.constraint(ExclusiveRoleConstraint(*roles))

    def ring(self, constraint_types, *roles):
        self.constraint(RingConstraint(constraint_types, *roles))

    def role_value_constraint(self, role: Role, values: Sequence[Union[T, Range[T]]]):
        first_type = values[0]._type() if isinstance(values[0], Range) else type(values[0])
        for v in values[1:]:
            t = v._type() if isinstance(v, Range) else type(v)
            if t != first_type:
                raise Exception("Values for role value constraint must have the same type.")
        if role.player()._is_primitive() or role.player()._is_value_type():
            self.constraint(RoleValueConstraint(role, values))
        else:
            raise Exception("A role value constraint can only be applied to roles played by value types")

    def exclusive_subtype_constraint(self, *concepts:Concept):
        self.constraint(ExclusiveSubtypeConstraint(*concepts))

    def inclusive_subtype_constraint(self, *concepts:Concept):
        self.constraint(InclusiveSubtypeConstraint(*concepts))

    def _add_concept(self, concept: Concept) -> Concept:
        self._validate_type_name(concept._name)
        if concept._name not in self.qb_model().concepts:
            self.qb_model().concepts[concept._name] = concept
        return concept

    def _ref_scheme_constraints(self, *relations:Relationship):
        if len(relations) == 1:
            # binary case, internal UC
            role = relations[0][1]
            self._internal_preferred_uc(role)
        else:
            roles = [rel[1] for rel in relations]
            self._composite_preferred_uc(*roles)

    def _internal_preferred_uc(self, role):
        rel = role._part_of()
        if rel._arity() != 2:
            raise Exception("The relationship should be binary to apply preferred identifier constraint")
        # mark the role as preferred identifier
        self.constraint(Unique(role, is_preferred_identifier=True))
        # mark the sibling role as mandatory and unique
        sibling = role.sibling()
        self.mandatory(sibling)
        self.unique(sibling)

    def _composite_preferred_uc(self, *roles):
        for role in roles:
            sibling = role.sibling()
            if not sibling:
                raise Exception("Composite preferred identifier constraint should be applied on binary relationships")
            self.mandatory(sibling)
            self.unique(sibling)
        self.constraint(Unique(*roles, is_preferred_identifier=True))

    def lookup_concept(self, name) -> Optional[qb.Concept]:
        name = to_pascal_case(name)
        if name in self._qb_model.concepts:
            return self._qb_model.concepts[name]
        return None

    def constraints(self):
        """Getter for the _constraints property"""
        return self._constraints

    def queries(self):
        return self._queries

    def concepts(self):
        return list(self._qb_model.concepts.values())

    def concepts_map(self):
        return self._qb_model.concepts

    def enums(self):
        return list(self._qb_model.enums.values())

    def enums_map(self):
        return self._qb_model.enums

    def value_types(self):
        return list(filter(lambda c: c._is_primitive(), self._qb_model.concepts.values()))

    def value_types_map(self):
        return dict(filter(lambda item: item[1]._is_primitive(), self._qb_model.concepts.items()))

    def entity_types(self):
        return list(filter(lambda c: not c._is_primitive(), self._qb_model.concepts.values()))

    def entity_types_map(self):
        return dict(filter(lambda item: not item[1]._is_primitive(), self._qb_model.concepts.items()))

    def relationships(self):
        return self._qb_model.relationships

    def bindable_tables(self):
        return self._bindable_tables

    def api(self):
        self._executor = self._executor or Executor(self._qb_model._config)
        return self._executor

    def table(self, name: str) -> SnowflakeTable:
        table = SnowflakeTable(name, self)
        self._bindable_tables[name] = table
        return table

    def csv_table(self, name: str, schema: dict[str, qb.Concept]) -> CsvTable:
        table = CsvTable(name, schema, self)
        self._bindable_tables[name] = table
        return table

    def binding(self, binding: Binding):
        self._bindings.append(binding)

    def generate_model_rules(self):
        if self._weaver is None:
            self._weaver = Weaver(self)
        else:
            raise Exception("Model rules have already been generated.")
        self._weaver.generate()

    def delete(self):
        self.api().provider().delete_model(self.name)

    def _validate_type_name(self, name):
        if name in self._qb_model.concepts:
            raise Exception(
                f"The name '{name}' is used to declare a Concept.")



from __future__ import annotations

import duckdb
from pandas import DataFrame
from typing import Any, Union

from .. import Compiler
from relationalai.early_access.metamodel import ir, executor as e, factory as f

class DuckDBExecutor(e.Executor):

    def __init__(self, skip_denormalization: bool = False) -> None:
        super().__init__()
        self.compiler = Compiler(skip_denormalization)

    def execute(self, model: ir.Model, task:ir.Task) -> Union[DataFrame, Any]:
        """ Execute the SQL query directly. """
        model = f.compute_model(f.logical([model.root, task]))

        connection = duckdb.connect()
        try:
            result = connection.query(self.compiler.compile(model)).to_df()
            return result
        finally:
            connection.close()


"""
Garbage collection pass that removes unused types and relations from the model.
"""
from dataclasses import dataclass, field

from relationalai.early_access.metamodel import ir, visitor, compiler
from relationalai.early_access.metamodel.util import FrozenOrderedSet

@dataclass
class GarbageCollectNodes(compiler.Pass):
    """
    A pass that removes unused types and relations from the model.
    """

    # TODO: Since sometimes models don't initially have the correct set of relations,
    # we sometimes actually end up /adding/ relations to the model here.

    # Flags to determine which nodes to gargbage collect
    # By default, we only garbage collect types and relations
    types: bool = field(default=True)
    relations: bool = field(default=True)

    @staticmethod
    def used_relations(model, engines) -> FrozenOrderedSet[ir.Relation]:
        return visitor.collect_by_type(ir.Relation, *engines, model.root, *model.annotations).frozen()

    @staticmethod
    def used_types(model, engines, relations) -> FrozenOrderedSet[ir.Type]:
        return visitor.collect_by_type(ir.Type, *engines, *relations, model.root, *model.annotations).frozen()

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        gc_types = self.types
        gc_relations = self.relations

        # No flags set, nothing to GC
        if not gc_relations and not gc_types:
            return model

        engines = model.engines
        relations = model.relations
        types = model.types

        # Note that engines and relations are mutually dependent. If we were to also GC engines,
        # then we would need to iterate until we reach a fixed point.
        if gc_relations:
            relations = self.used_relations(model, engines)

        # Types are independent of engines and relations, so we can just collect them last
        if gc_types:
            types = self.used_types(model, engines, relations)

        return ir.Model(
            engines,
            relations,
            types,
            model.root,
            model.annotations,
        )



import threading

from relationalai.early_access.dsl.orm.constraints import Mandatory, Unique
from relationalai.early_access.dsl.orm.models import Role, Relationship
from relationalai.early_access.metamodel.util import OrderedSet
from relationalai.early_access.dsl.core import warn

_reasoner_local = threading.local()

def init_reasoner(model):
    _reasoner_local.instance = Reasoner(model)

def get_reasoner():
    try:
        return _reasoner_local.instance
    except AttributeError:
        raise RuntimeError("Reasoner not initialized. Call `init_reasoner(model)` before using it.")

class Reasoner:

    def __init__(self, model):
        self._model = model

        # Map each role's GUID to the relationship it belongs to
        self._role_part_of = {
            role._guid(): relationship
            for relationship in model.relationships()
            for role in (relationship[i] for i in range(relationship._arity()))
        }

        self._role_mandatory = {}
        self._role_spanned_by = {}

        # Process constraints
        for constraint in model._constraints:
            if isinstance(constraint, Mandatory):
                roles = constraint.roles()
                if len(roles) == 1:
                    self._role_mandatory[roles[0]._guid()] = constraint

            elif isinstance(constraint, Unique):
                for r in constraint.roles():
                    self._role_spanned_by[r._guid()] = constraint

        # Filter out external unique constraints
        self._role_spanned_by_internal = {}
        for guid, uc in self._role_spanned_by.items():
            role_guids = {r._part_of()._guid() for r in uc._roles}
            if len(role_guids) == 1:
                self._role_spanned_by_internal[guid] = uc

    # [REKS: TODO] Need to loosen this requirement so that we admit
    #              compatible roles when the player of r1 is a subtype
    #              of that of r2 or vice versa.
    def compatible(self, r1, r2):
        warn(f"Checking compatibility of roles played by {str(r1.player())} and {str(r2.player())} without reasoning about subtypes")
        return r1.player() == r2.player()

    # [REKS: TODO] Need to do the subtype/supertype reasoning required
    #              to properly answer this w/o requring c1 to be equal
    #              to c2
    def least_supertype(self, c1, c2):
        warn(f"Trying to find least supertpe or {str(c1)} and {str(c2)} without reasoning about subtypes")
        return None if str(c1) != str(c2) else c1

    def mandatory(self, role) -> bool:
        return role._guid() in self._role_mandatory

    def model(self): return self._model

    # A role is a "one" role (as opposed to a "many" role) if it is not
    #   spanned by an internal uniqueness constraint
    def one_role(self, role) -> bool:
        return False if role._guid() in self._role_spanned_by_internal else True

    def part_of(self, role) -> Relationship:
        return self._role_part_of[role._guid()]
    
    # [REKS: TODO] Notice that we retain the sibling role ordering from
    #              that of the Relationship, which we currently use to
    #              generate atoms in class MeasureRule and its subclasses.
    #              However, this only works if the measure role of the
    #              Relationship appears in the last position -- which is
    #              almost always the case but is not guaranteed, and we
    #              don't check for it currently
    #
    def role_siblings(self, role) -> OrderedSet[Role]:
        role_guid = role._guid()
        relationship = self.part_of(role)

        return OrderedSet.from_iterable([
            relationship[i]
            for i in range(relationship._arity())
            if relationship[i]._guid() != role_guid
        ])


from dataclasses import dataclass
from enum import Enum
from typing import Optional, Any


@dataclass
class ORMEntityType:
    id: str
    name: str
    ref_mode: Optional[str]


@dataclass
class ORMValueType:
    id: str
    name: str
    data_type: str


@dataclass
class ORMRole:
    id: str
    name: str
    relationship_name: str
    player: str


@dataclass
class ORMSubtypeFact:
    subtype_role_id: Optional[str]
    subtype: Optional[str]
    supertype_role_id: Optional[str]
    supertype: Optional[str]


@dataclass
class SubtypeFact:
    subtype_name: str
    supertype_name: str


@dataclass
class ExclusiveSubtypeFact(SubtypeFact):
    pass


@dataclass
class ExclusiveInclusiveSubtypeFact(SubtypeFact):
    pass


@dataclass
class ORMUniquenessConstraint:
    id: str
    roles: list[str]
    identifies: Optional[str]


@dataclass
class ORMMandatoryConstraint:
    id: str
    roles: list[str]


@dataclass
class ORMRingConstraint:
    id: str
    roles: list[str]
    ring_types: list['ORMRingType']


class ORMRingType(Enum):
    IRREFLEXIVE = 'Irreflexive'
    ANTISYMMETRIC = 'Antisymmetric'
    ASYMMETRIC = 'Asymmetric'
    STRONGLY_INTRANSITIVE = 'StronglyIntransitive'
    INTRANSITIVE = 'Intransitive'
    ACYCLIC = 'Acyclic'
    PURELY_REFLEXIVE = 'PurelyReflexive'
    REFLEXIVE = 'Reflexive'
    SYMMETRIC = 'Symmetric'
    TRANSITIVE = 'Transitive'


@dataclass
class ORMExclusionConstraint:
    id: str
    roles: list[str]
    inclusive: bool


@dataclass
class ORMInclusionConstraint:
    id: str
    roles: list[str]
    exclusive: bool


@dataclass
class ORMReadingOrder:
    roles: list['ORMReadingOrderRole']


@dataclass
class ORMReadingOrderRole:
    index: int
    prefix: Optional[str]
    postfix: Optional[str]
    text: Optional[str]
    role: Optional[ORMRole]


@dataclass
class ORMRoleValueConstraint:
    role: ORMRole
    values: list[Any]


@dataclass
class RoleConstraint:
    roles: list[str]
    inclusive: bool
    exclusive: bool



"""
A simple metamodel for SQL.
"""
from __future__ import annotations
from dataclasses import dataclass
from io import StringIO
from typing import Any, Optional, Tuple, Union
from relationalai.early_access.metamodel.util import Printer as BasePrinter
import json


@dataclass(frozen=True)
class Node:
    def __str__(self):
        return to_string(self)

@dataclass(frozen=True)
class Program(Node):
    """ The root node of the SQL program. """
    statements: list[Node]

#-------------------------------------------------
# Table
#-------------------------------------------------

@dataclass(frozen=True)
class Column(Node):
    name: str
    type: str

@dataclass(frozen=True)
class Table(Node):
    name: str
    columns: list[Column]

#-------------------------------------------------
# Statements
#-------------------------------------------------

@dataclass(frozen=True)
class CreateTable(Node):
    table: Table

@dataclass(frozen=True)
class CreateView(Node):
    name: str
    query: Union[Select, CTE]

@dataclass(frozen=True)
class Insert(Node):
    table: str
    columns: list[str]
    values: list[Tuple[Any, ...]]
    select: Optional[Select]

@dataclass(frozen=True)
class Update(Node):
    table: str
    set: list[UpdateSet]
    where: Optional[Where]

@dataclass(frozen=True)
class UpdateSet(Node):
    name: str
    expression: str

@dataclass(frozen=True)
class Select(Node):
    distinct: bool
    vars: list[Union[VarRef, RowNumberVar, int]]
    froms: list[From]
    where: Optional[Where]
    group_by: Optional[list[VarRef]] = None

@dataclass(frozen=True)
class VarRef(Node):
    name: str
    column: Optional[str]
    alias: Optional[str]

@dataclass(frozen=True)
class From(Node):
    table: str
    alias: Optional[str]

@dataclass(frozen=True)
class RowNumberVar(Node):
    order_by_vars: list[OrderByVar]
    partition_by_vars: list[str]
    alias: str

@dataclass(frozen=True)
class OrderByVar(Node):
    var: str
    is_ascending: bool

# TODO: consider removing Where and make Select.where: Expr
@dataclass(frozen=True)
class Where(Node):
    expression: Expr

@dataclass(frozen=True)
class RawSource(Node):
    """ A raw source for SQL, used to represent a raw SQL query or expression. """
    src: str

#-------------------------------------------------
# Clauses
#-------------------------------------------------
# TODO: move clauses from other sections

@dataclass(frozen=True)
class CTE(Node):
    """ Common Table Expressions. """
    recursive: bool
    name: str
    columns: list[str]
    selects: list[Select]

#-------------------------------------------------
# Expressions
#-------------------------------------------------

@dataclass(frozen=True)
class Expr(Node):
    pass

@dataclass(frozen=True)
class And(Expr):
    expr: list[Expr]

@dataclass(frozen=True)
class Or(Expr):
    expr: list[Expr]

@dataclass(frozen=True)
class Terminal(Expr):
    """ Avoid going deeper in the meta-model, this is an arbitrary terminal expression."""
    expr: str

@dataclass(frozen=True)
class NotNull(Expr):
    var: str

@dataclass(frozen=True)
class NotExists(Expr):
    expr: Select

@dataclass(frozen=True)
class Like(Expr):
    left: str
    right: str

#--------------------------------------------------
# Printer
#--------------------------------------------------

def to_string(node) -> str:
    io = StringIO()
    Printer(io).print_node(node, 0)
    return io.getvalue()

@dataclass(frozen=True)
class Printer(BasePrinter):

    def _join(self, args, sep=', ', indent=0):
        for i, s in enumerate(args):
            if i != 0:
                if indent != 0:
                    self._print(sep.rstrip())
                    self._nl()
                else:
                    self._print(sep)
            self.print_node(s, indent)

    def _print_value(self, value, quote_strings:bool = False):
        if isinstance(value, tuple):
            self._print("(")
            for i, v in enumerate(value):
                if i != 0:
                    self._print(", ")
                self._print_value(v, True)
            self._print(")")
        elif isinstance(value, str):
            if quote_strings:
                self._print(json.dumps(value))
            else:
                self._print(value)
        elif isinstance(value, bool):
            self._print(str(value).lower())
        else:
            self._print(str(value))

    def print_node(self, node:Node, indent=0, inlined=False) -> None:
        # Table
        if isinstance(node, Program):
            for idx, d in enumerate(node.statements):
                self.print_node(d, indent)
                # Avoid an extra newline at the end of the file.
                if idx != len(node.statements) - 1:
                    self._nl()
        elif isinstance(node, Table):
            self._print(f"TABLE {node.name} (")
            self._join(node.columns)
            self._print(")")
        elif isinstance(node, Column):
            self._print(f"{node.name} {node.type}")

        # Statements
        elif isinstance(node, CreateTable):
            self._print(f"CREATE {node.table};")
        elif isinstance(node, CreateView):
            # TODO - crying a bit inside :(
            self._print_nl(f"DROP TABLE IF EXISTS {node.name};")
            self._print(f"CREATE VIEW {node.name} AS ")
            self.print_node(node.query, indent, True)
            self._print(";")
        elif isinstance(node, Insert):
            self._print(f"INSERT INTO {node.table} ")
            if len(node.columns) > 0:
                self._print("(")
                self._join(node.columns)
                self._print(") ")
            if len(node.values) > 0:
                self._print("VALUES (")
                self._join(node.values)
                self._print(")")
            if node.select is not None:
                self.print_node(node.select, indent, True)
            self._print(";")
        elif isinstance(node, Update):
            self._print(f"UPDATE {node.table} SET ")
            self._join(node.set)
            if node.where is not None:
                self.print_node(node.where, indent)
            self._print(";")
        elif isinstance(node, UpdateSet):
            self._print(f"{node.name} = {node.expression}")
        elif isinstance(node, Select):
            self._indent_print(indent, "SELECT ")
            if node.distinct:
                self._print("DISTINCT ")
            self._join(node.vars)
            self._print(" FROM ")
            self._join(node.froms)
            if node.where:
                self.print_node(node.where, indent)
            if node.group_by:
                self._print(" GROUP BY ")
                self._join(node.group_by)
            if not inlined:
                self._print(";")
        elif isinstance(node, VarRef):
            if node.column is None:
                self._print(node.name)
            else:
                self._print(f"{node.name}.{node.column}")
            if node.alias and node.alias != node.column:
                self._print(f" as {node.alias}")
        elif isinstance(node, RowNumberVar):
            self._print("ROW_NUMBER() OVER (")
            if node.partition_by_vars:
                self._print(" PARTITION BY ")
                self._join(node.partition_by_vars)
            self._print(" ORDER BY ")
            self._join(node.order_by_vars)
            self._print(f" ) as {node.alias}")
        elif isinstance(node, OrderByVar):
            self._print(node.var)
            if node.is_ascending:
                self._print(" ASC")
            else:
                self._print(" DESC")
        elif isinstance(node, From):
            self._print(node.table)
            if node.alias is not None:
                self._print(f" AS {node.alias}")
        elif isinstance(node, Where):
            self._print(f" WHERE {node.expression}")
        elif isinstance(node, RawSource):
            self._print(node.src)

        # Clauses
        elif isinstance(node, CTE):
            self._print("WITH ")
            if node.recursive:
                self._print("RECURSIVE ")
            self._print(f"{node.name} (")
            self._join(node.columns)
            self._print_nl(") AS (")
            for i, s in enumerate(node.selects):
                if i != 0:
                    self._nl()
                    self._indent_print_nl(indent + 2, "UNION ALL")
                self.print_node(s, indent + 1, True)
            self._nl()
            self._print(f") SELECT * FROM {node.name}")
            if not inlined:
                self._print(";")
        # --------------------------------------------------
        # Primitives, Annotations
        # --------------------------------------------------
        elif isinstance(node, (str, int, float, bool, tuple)):
            self._indent_print(indent, "")
            self._print_value(node)
        # Expressions
        elif isinstance(node, And):
            self._join(node.expr, " AND ", indent)
        elif isinstance(node, Or):
            self._join(node.expr, " OR ", indent)
        elif isinstance(node, Terminal):
            self._print(node.expr)
        elif isinstance(node, NotNull):
            self._print(f"{node.var} IS NOT NULL")
        elif isinstance(node, NotExists):
            self._print("NOT EXISTS ( ")
            self.print_node(node.expr, indent, True)
            self._print(" )")
        elif isinstance(node, Like):
            self._print(f"{node.left} LIKE {node.right}")
        else:
            raise Exception(f"Missing SQL.pprint({type(node)}): {node}")


from .. import dsl
from ..std.graphs import unwrap
from ..std.graphs import Graph
from ..std import minimum, maximum
from enum import Enum


# --------------------------------------------------
# Ego network
# --------------------------------------------------
def ego_network(graph, node, hops):
    if not isinstance(hops, int):
        raise TypeError(f"'hops' must be an integer, got {type(hops).__name__} instead")
    elif hops < 0:
        raise ValueError(f"'hops' must be non-negative, got {hops}")

    a, b = dsl.create_vars(2)
    dsl.global_ns.graphlib_experimental.ego_network(graph, unwrap(node), hops, a, b)
    la = graph.compute._lookup(a)
    lb = graph.compute._lookup(b)
    return (la, lb)

# --------------------------------------------------
# to_undirected_graph
# --------------------------------------------------

class AggregationFunction(Enum):
    SUM = 0
    MIN = 1
    MAX = 2
_Agg = AggregationFunction

def to_undirected_graph(model, in_graph:Graph, attributes=[], *, aggregation:_Agg):
    # This expects a directed graph that does not have multi-edges in Set semantics.
    # Examples
    # The edges  (1, 2, 1.0) and (1, 2, 1.0) are allowed but will be counted once.
    # The edges  (1, 2, 1.0) and (2, 1, 1.0) are allowed, and will
    #   be transformed into one undirected edge between nodes 1 and 2 with weight
    #   2.0 (if the aggregation is sum), or 1.0 (if the aggregation is max or min).
    # The edges  (1, 2, 1.0) and (2, 1, 2.0) are allowed, and will
    #   be transformed into one undirected edge between nodes 1 and 2 with weight
    #   3.0 (if the aggregation is sum), or 1.0 (if the aggregation is min)
    #   or 2.0 (if the aggregation is max).
    # The edges  (1, 2, 2.0) and (2, 1, 1.0) and (2, 1, 1.0) are allowed, and will
    #   be transformed into one undirected edge between nodes 1 and 2 with weight
    #   3.0 (if the aggregation is sum), or 1.0 (if the aggregation is min)
    #   or 2.0 (if the aggregation is max).
    assert not in_graph.undirected, f"Expect directed input graph. Got undirected input `graph{in_graph.id}`."

    # Create graph
    out_graph = Graph(model, undirected = True, weighted = in_graph.weighted)
    OutNode, OutEdge = out_graph.Node, out_graph.Edge
    InNode, InEdge = in_graph.Node, in_graph.Edge

    # We get automatic error detection if we pass an attribute that does not exist.
    # And if a certain node does not have an attribute, and others have that attribute,
    # it will be set to null (Rel takes care of this).
    OutNode.extend(InNode)
    with model.rule():
        n = InNode()
        OutNode(n).set(**{attr: getattr(n, attr) for attr in attributes})

    if in_graph.weighted:

        # E(a,b) and not E(b,a)
        with model.rule():
            e1 = InEdge()
            with model.not_found():
                InEdge(from_ = e1.to, to = e1.from_)
            OutEdge.add(from_ = e1.from_, to = e1.to, weight = e1.weight)

        # E(a,b) and E(b,a)
        with model.rule():
            e1 = InEdge()
            # This is a useful performance improvement, since we cut
            # the size of the produced edge-list in half. Additionally, it
            # excludes self loops from the aggregation, which would be
            # incorrect. Self-loops are handled separately, below.
            e1.from_ < e1.to
            e2 = InEdge(from_ = e1.to, to = e1.from_)
            if aggregation == _Agg.SUM:
                OutEdge.add(from_ = e1.from_, to = e1.to, weight = e1.weight + e2.weight)
            elif aggregation == _Agg.MIN:
                OutEdge.add(from_ = e1.from_, to = e1.to, weight = minimum(e1.weight, e2.weight))
            elif aggregation == _Agg.MAX:
                OutEdge.add(from_ = e1.from_, to = e1.to, weight = maximum(e1.weight, e2.weight))

        # E(a,a)
        with model.rule():
            e1 = InEdge()
            e1.from_ == e1.to
            OutEdge.add(from_ = e1.from_, to = e1.to, weight = e1.weight)

    else:  # Unweighted input

        with model.rule():
            e = InEdge()
            OutEdge.add(from_ = e.from_, to = e.to)

    return out_graph



import hashlib

def hash_string(s: str, digits=32):
    return hashlib.md5(s.encode()).hexdigest()[:digits]

def database_name_from_sproc_name(sproc_name: str) -> str:
    # Import here to avoid circular import issues
    from relationalai.clients.util import IdentityParser, ParseError
    from relationalai.rel_utils import sanitize_identifier

    err_text = f"Invalid stored procedure name: {sproc_name}"
    try:
        parser = IdentityParser(sproc_name, require_all_parts=True)
        if not (parser.identity and parser.db and parser.schema and parser.entity):
            raise ValueError(err_text)

        # Remove all special characters from the name and replace with '_'
        sanitized_name = sanitize_identifier(parser.identity)
        hashed_sanitized_name = hash_string(sanitized_name)

        # If the name is less than 32 characters, append the hash of the name
        if len(sanitized_name) < 32:
            return f"{sanitized_name}_{hashed_sanitized_name}"

        # If the name is more than 32 characters, truncate it to 31 characters and sanitize each part
        truncated_database = sanitize_identifier(parser.db[:10])
        truncated_schema = sanitize_identifier(parser.schema[:10])
        truncated_proc = sanitize_identifier(parser.entity[:(31 - len(truncated_database) - len(truncated_schema))])
        return f"{truncated_database}_{truncated_schema}_{truncated_proc}_{hashed_sanitized_name}"
    except ParseError:
        raise ValueError(err_text)



from __future__ import annotations
from typing import Any, Union
from decimal import Decimal as PyDecimal

from .. import builder as b

_Number = Union[b.Producer, float, int, PyDecimal]
_Integer = Union[b.Producer, int]

def _make_expr(op: str, *args: Any) -> b.Expression:
    return b.Expression(b.Relationship.builtins[op], *args)

def abs(value: _Number) -> b.Expression:
    return _make_expr("abs", value, b.Number.ref("res"))

def natural_log(value: _Number) -> b.Expression:
    return _make_expr("natural_log", value, b.Number.ref("res"))



import os
from typing import List, Set, Tuple, Union
import zipfile
from enum import Enum

from relationalai import dsl
from relationalai.dsl import Graph, Instance, Type, Property, create_vars, get_graph, next_id
from relationalai.metamodel import ActionType, Builtins
from relationalai.std import rel

class VarLength:
    def __init__(self, type: Type):
        self.type = type
        self.min = 1
        self.max = 1

    def __getitem__(self, key):
        if isinstance(key, slice):
            self.min = key.start
            self.max = key.stop
        else:
            raise ValueError("VarLength only supports slicing")
        return self

class PathSelection(Enum):
    ALL = "all" # return all paths between source and target
    SINGLE = "single" # return a single path between source and target
    LIMIT = "limit" # return a specified number paths between source and target, as given by parameter k.
    RANDOM = "random" # return a single randomly-chosen path between source and target, seeded with parameter seed.
    UNIFORM_RANDOM = "uniform_random" # return a single uniform randomly-chosen path between source and target, seeded with parameter seed.

class PathGroup(Enum):
    ANY = "any"
    FOREACH = "for_each"

class Strategy(Enum):
    FROM_SOURCE = "from_source"
    TWO_SIDED = "two_sided"

class PathQuery:

    def __init__(
        self,
        model: Graph,
        segments: List[Union[Instance,VarLength]],
        backwards_edge: str,
        group: PathGroup = PathGroup.FOREACH,
        selection: PathSelection = PathSelection.ALL,
        strategy: Strategy = Strategy.FROM_SOURCE,
    ):
        # create types and instance
        self._path_id = next_id()
        self._conn_relation_name = f"conn_{self._path_id}"
        self._model = model

        # create types
        self._Path = dsl.Type(model, f"Path{self._path_id}", omit_intrinsic_type_in_hash=True)
        self._PathNode = dsl.Type(model, f"PathNode{self._path_id}", omit_intrinsic_type_in_hash=True)
        self._PathEdge = dsl.Type(model, f"PathEdge{self._path_id}", omit_intrinsic_type_in_hash=True)

        # declare multivalued attributes
        self._Path.nodes.has_many()
        self._Path.edges.has_many()

        # mark these types and their attributes as @no_inline
        self._add_noinline(self._Path, [])
        self._add_noinline(self._PathNode, ["path", "index", "node_id"])
        self._add_noinline(self._PathEdge, ["path", "index", "label"])

        # create instance
        self._instance = self._Path()
        self._match_called = False

        # store path query attributes
        self._segments = segments
        self._backwards_edge = backwards_edge
        self._group = group
        self._selection = selection
        self._strategy = strategy

    def _add_noinline(self, typ, props):
        typ._type.parents.append(Builtins.NoInlineAnnotation)
        for prop in props:
            getattr(typ, prop)._prop.parents.append(Builtins.NoInlineAnnotation)

    def _match(self):
        # avoid emitting the Rel multiple times, which throws off Pathfinder
        if not self._match_called:
            self._match_called = True

            begin_var, end_var = self._match_inner()

            self._begin_var = begin_var
            self._end_var = end_var

        return self._conn_relation_name

    def _match_inner(self):
        begin_var = None
        cur_end = None

        for idx, segment in enumerate(self._segments):
            if isinstance(segment, VarLength):
                var_length_rel_name = self._var_length(idx, segment)
                seg_begin, seg_end = create_vars(2)
                getattr(rel, var_length_rel_name)(seg_begin, seg_end)

                if begin_var is None:
                    begin_var = seg_begin
                    cur_end = seg_end
                else:
                    getattr(seg_begin, self._backwards_edge) == cur_end
                    cur_end = seg_end
            else:
                next_var = segment
                if begin_var is None:
                    begin_var = next_var
                    cur_end = begin_var
                else:
                    getattr(next_var, self._backwards_edge) == cur_end
                    cur_end = next_var

        conn_relation = getattr(rel, self._conn_relation_name)
        conn_relation._rel.parents.append(Builtins.NoInlineAnnotation)
        conn_relation.add(begin_var, cur_end)

        return begin_var, cur_end

    def _var_length(self, segment_idx: int, segment: VarLength):
        # relations for each length
        prefix = f"conn_{self._path_id}_segment_{segment_idx}"

        lines = []

        for length in range(1, segment.max+1):
            length_rel_name = f"{prefix}_length_{length}"

            if length == 1:
                lines.extend([
                    f"def {length_rel_name}(a, a1):",
                    f"    {segment.type._type.name}(a) and",
                    "    a = a1"
                ])
                # would just project `(a, a)`, but that gives a
                # bunch of Rel warnings`

            else:
                prev_length_relation_name = f"{prefix}_length_{length-1}"
                lines.extend([
                    f"def {length_rel_name}(a, c):",
                    "    exists((b) |",
                    f"       {prev_length_relation_name}(a, b) and",
                    f"       {segment.type._type.name}(c) and",
                    f"       {self._backwards_edge}(c, b)",
                    "     )"
                ])

        # union them all together in overall segment relation
        lines.extend([
            f"def {prefix}(start, finish):",
            "    " + " or\n    ".join([
                f"{prefix}_length_{length}(start, finish)"
                for length in range(1, segment.max+1)
            ]),
        ])

        self._model.install_raw(
            '\n'.join(lines),
            name=f"path_{self._path_id}_segment_{segment_idx}",
        )

        return prefix


class PathInstance(Instance):

    def __init__(
        self,
        path_query: PathQuery,
    ):
        self._path_query = path_query

        self._already_called = False
        super().__init__(
            self._path_query._model,
            ActionType.Get, [self._path_query._instance],
            named={},
        )

    # implements producer API
    def _to_var(self):
        self._invoke_pathfinder()
        return super()._to_var()

    def _invoke_pathfinder(self):
        if self._already_called:
            return
        self._already_called = True

        # ensure that the pathfinder Rel library is installed
        _install_pathfinder(self._path_query._model)

        source_rel = f"source_{self._path_query._path_id}"
        target_rel = f"target_{self._path_query._path_id}"
        getattr(rel, source_rel).add(self._path_query._begin_var)
        getattr(rel, target_rel).add(self._path_query._end_var)

        shortest_paths_rel = f"shortest_paths_{self._path_query._path_id}"
        path_edge_rel = f"path_edge_{self._path_query._path_id}"
        path_node_rel = f"path_node_{self._path_query._path_id}"

        invocation = f"""
        @no_inline
        def {shortest_paths_rel} {{
            ::pathfinder::shortest_paths[
                :{self._path_query._group.value},
                :{self._path_query._selection.value},
                :{self._path_query._strategy.value},
                {self._path_query._conn_relation_name},
                {source_rel},
                {target_rel}
            ]
        }}

        @no_inline
        def {path_node_rel}(path, index, id):
            {shortest_paths_rel}(path, :node, index, id)

        @no_inline
        def {path_edge_rel}(path, index, label):
            {shortest_paths_rel}(path, :edge_label, index, label)
        """

        self._path_query._model.install_raw(invocation)

        # select out nodes
        with self._path_query._model.rule():
            path_id, index, node_label = create_vars(3)
            getattr(rel, path_node_rel)(path_id, index, node_label)
            path = self._path_query._Path.add(id=path_id)
            # TODO: avoid perf hit of going in both directions (path<->node)
            node = self._path_query._PathNode.add(path=path, index=index, value=node_label)
            path.nodes.add(node)

        # select out edges
        with self._path_query._model.rule():
            path_id, index, label = create_vars(3)
            getattr(rel, path_edge_rel)(path_id, index, label)
            path = self._path_query._Path.add(id=path_id)
            # TODO: avoid perf hit of going in both directions (path<->edge)
            edge = self._path_query._PathEdge.add(path=path, index=index, label=label)
            path.edges.add(edge)

# TODO: automatically infer this from the path query itself
def enable_on(edge: Property, filter_attrs: List[Tuple[Type, Set[str]]]):  # noqa: F821
    edge._prop.parents.append(Builtins.PQEdgeAnnotation)
    edge._prop.parents.append(Builtins.NoInlineAnnotation)
    for type, attrs in filter_attrs:
        type._type.parents.append(Builtins.PQFilterAnnotation)
        type._type.parents.append(Builtins.NoInlineAnnotation)
        for attr_name in attrs:
            prop = getattr(type, attr_name)
            prop._prop.parents.append(Builtins.PQFilterAnnotation)
            prop._prop.parents.append(Builtins.NoInlineAnnotation)

# main entry point
def path(
    segments: List[Union[Instance,VarLength]],
    backwards_edge: str,
    group: PathGroup = PathGroup.FOREACH,
    selection: PathSelection = PathSelection.ALL,
    strategy: Strategy = Strategy.FROM_SOURCE,
):
    """
    Find a path consisting of the provided segments, linked by
    the provided `backwards_edge` property.
    """

    model = get_graph()

    query = PathQuery(model, segments, backwards_edge, group, selection, strategy)

    # declare filter relations
    conn_relation_name = query._match()

    # invoke filter
    a, b = create_vars(2)
    getattr(rel, conn_relation_name)(a, b)

    # return a special Producer representing the path itself, which
    # lazily invokes Pathfinder
    return PathInstance(query)

def _get_pathfinder_source():
    # check if the current dir is in a zip file
    current_dir = os.path.dirname(__file__)
    pathfinder_path = os.path.join(current_dir, "pathfinder.rel")
    if os.path.exists(pathfinder_path):
        return open(pathfinder_path).read()
    # if we're in a zip file, read it within that
    zip_split = current_dir.split(".zip")
    zip_path = zip_split[0] + ".zip"
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        with zip_ref.open('relationalai/experimental/paths/pathfinder.rel') as file:
            return file.read().decode("utf-8")

def _install_pathfinder(model: Graph):
    source = _get_pathfinder_source()
    model.install_raw(source, name="pathfinder", overwrite=True)



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from typing import cast, Optional

from relationalai.early_access.metamodel import ir, compiler as c, factory as f
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
from relationalai.early_access.metamodel.visitor import Visitor


class RecursiveUnion(c.Pass):
    """
        Pass to rewrite recursive relations as unions of `ir.Logical` nodes.

        This pass identifies recursive relations by detecting `ir.Logical` nodes that both read from and write to the same `ir.Relation`.
        Once a recursive relation is found, all logicals that write to that relation are grouped together.
        These groups are then wrapped in an `ir.Union`, and the model's root task is reconstructed accordingly.
    """

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        rpv = TopLogicalReadWriteVisitor()
        model.accept(rpv)

        reads: dict[ir.Logical, OrderedSet[int]] = rpv.reads()
        writes: dict[ir.Logical, OrderedSet[int]] = rpv.writes()

        # Step 1: Identify recursive relation IDs (read & write in same logical)
        recursive_rel_ids = {
            rel_id
            for logical, write_ids in writes.items()
            for rel_id in write_ids & reads.get(logical, OrderedSet())
        }

        # Step 2: Group ir.Logical nodes that write to any recursive relation
        recursive_groups: dict[int, OrderedSet[ir.Logical]] = defaultdict(OrderedSet)
        recursive_logicals: OrderedSet[ir.Logical] = OrderedSet()

        for logical, write_ids in writes.items():
            for rel_id in write_ids:
                if rel_id in recursive_rel_ids:
                    recursive_groups[rel_id].add(logical)
                    recursive_logicals.add(logical)

        if recursive_logicals:

            # Step 3: Construct the new logical body
            root_logical = cast(ir.Logical, model.root)
            new_body = [logical for logical in root_logical.body if logical not in recursive_logicals]

            # Step 4: Add unions for each recursive group
            for group in recursive_groups.values():
                if group:
                    new_body.append(f.union(list(group)))

            return model.reconstruct(model.engines, model.relations, model.types, f.logical(new_body), model.annotations)

        return model

@dataclass
class TopLogicalReadWriteVisitor(Visitor):
    """
    Compute the set of reads and writes relation ids for top level Logical nodes.
    Skip unions, because we are using this pass to find recursive rule and union them.

    Note that reads are Lookups and writes are Updates. We don't consider Output a write
    because it is not targeting a relation.

    This visitor can be called from ir.Model or model root task.
    """

    _reads: dict[ir.Logical, OrderedSet[int]] = field(default_factory=dict)
    _writes: dict[ir.Logical, OrderedSet[int]] = field(default_factory=dict)

    def reads(self):
        return self._reads

    def writes(self):
        return self._writes

    _stack: list[ir.Logical] = field(default_factory=list)

    def visit_logical(self, node: ir.Logical, parent: Optional[ir.Node]):
        # We only track reads/writes for top-level Logical nodes and also skip top level program Logical node.
        if parent != ir.Model and len(self._stack) == 0:
            self._stack.append(node)
        super().visit_logical(node, parent)
        if self._stack:
            self._stack.pop()

    def visit_lookup(self, node: ir.Lookup, parent: Optional[ir.Node]):
        for logical in self._stack:
            if logical.id not in self._reads:
                self._reads[logical] = ordered_set()
            self._reads[logical].add(node.relation.id)
        return super().visit_lookup(node, parent)

    def visit_update(self, node: ir.Update, parent: Optional[ir.Node]):
        for logical in self._stack:
            if logical.id not in self._reads:
                self._writes[logical] = ordered_set()
            self._writes[logical].add(node.relation.id)
        return super().visit_update(node, parent)

    def visit_union(self, node: ir.Union, parent: Optional[ir.Node]):
        # Skip unions, because we are using this pass to find recursive rule and union them.
        pass


from __future__ import annotations
import random
import os
from typing import Optional

import snowflake.connector

import relationalai as rai
from relationalai import debugging
from relationalai.debugging import logger
from relationalai.clients import config as cfg
from relationalai.util.otel_handler import enable_otel_export
from relationalai.tools.constants import USE_PACKAGE_MANAGER
from relationalai.util.snowflake_handler import SnowflakeHandler
from relationalai.util.span_format_test import SpanCollectorHandler, assert_valid_span_structure
from relationalai.util.span_tracker import TRACE_ID, get_root_span_attrs

def graph_index_config_fixture():
    cloud_provider = os.getenv("RAI_CLOUD_PROVIDER")
    if cloud_provider:
        config = make_config()
    else:
        config = cfg.Config()

    config.set("use_graph_index", True)
    config.set("reuse_model", False)

    yield config
    return

def engine_config_fixture(size, use_direct_access=False, reset_session=False):
    # Check for an externally provided engine name
    # It is used in GitHub Actions to run tests against a specific engine
    engine_name = os.getenv("ENGINE_NAME")
    if engine_name:
        # If engine name was provided, just yield the config
        config = make_config(engine_name, use_direct_access=use_direct_access)
        # Try to reset the session instead of using active session if reset_session is true
        rai.Resources(config=config, reset_session=reset_session)

        if os.getenv("DISABLE_PROFILE_POLLING"):
            print('disabling `debug` in config')
            config.set("debug", False)

        config.set("reuse_model", False)
        config.set("enable_otel_handler", True) 

        yield config
        return

    # If there's a local config file, use it, including
    # the engine specified there.
    config = cfg.Config()
    if config.file_path is not None:
        # Try to reset the session instead of using active session if reset_session is true
        rai.Resources(config=config, reset_session=reset_session)
        yield config
        return

    # Otherwise, create a new engine and delete it afterwards.
    random_number = random.randint(1000000000, 9999999999)
    engine_name = f"pyrel_test_{random_number}"
    create_engine(engine_name, size=size, use_direct_access=use_direct_access)

    yield make_config(engine_name, use_direct_access=use_direct_access)

    delete_engine(engine_name, use_direct_access=use_direct_access)

def create_engine(engine_name: str, size: str,  use_direct_access=False):
    print('create_engine: about to call make_config')
    config = make_config(engine_name, use_direct_access=use_direct_access)

    provider = rai.Resources(config=config)
    print(f"Creating engine {engine_name}")
    provider.create_engine(name=engine_name, size=size)
    print(f"Engine {engine_name} created")

def delete_engine(engine_name: str, use_direct_access=False):
    print(f"Deleting engine {engine_name}")
    config = make_config(engine_name, use_direct_access=use_direct_access)
    provider = rai.Resources(config=config)
    provider.delete_engine(engine_name)
    print(f"Engine {engine_name} deleted")

def make_config(engine_name: str | None = None, fetch_profile: bool = True, use_package_manager = USE_PACKAGE_MANAGER, use_direct_access = False) -> cfg.Config:
    cloud_provider = os.getenv("RAI_CLOUD_PROVIDER")

    print('cloud provider:', cloud_provider)

    if cloud_provider is None:
        raise ValueError("RAI_CLOUD_PROVIDER must be set")
    elif cloud_provider == "azure":
        client_id = os.getenv("RAI_CLIENT_ID")
        client_secret = os.getenv("RAI_CLIENT_SECRET")
        if client_id is None or client_secret is None:
            raise ValueError(
                "RAI_CLIENT_ID, RAI_CLIENT_SECRET must be set if RAI_CLOUD_PROVIDER is set to 'azure'"
            )

        # Pull from env vars; Default to prod
        host = os.getenv("RAI_AZURE_HOST") or "azure.relationalai.com"
        creds_url = os.getenv("RAI_AZURE_CLIENT_CREDENTIALS_URL") or "https://login.relationalai.com/oauth/token"
        region = os.getenv("RAI_AZURE_REGION") or "us-east"

        return cfg.Config(

            {
                "platform": "azure",
                "host": host,
                "port": "443",
                "region": region,
                "scheme": "https",
                "client_credentials_url": creds_url,
                "client_id": client_id,
                "client_secret": client_secret,
                "engine": engine_name,
                "use_package_manager": use_package_manager,
            }
        )

    elif cloud_provider == "snowflake":
        sf_username = os.getenv("SF_TEST_ACCOUNT_USERNAME")
        sf_password = os.getenv("SF_TEST_ACCOUNT_PASSWORD")
        sf_account = os.getenv("SF_TEST_ACCOUNT_NAME")
        sf_role = os.getenv("SF_TEST_ROLE_NAME", "RAI_USER")
        sf_warehouse = os.getenv("SF_TEST_WAREHOUSE_NAME")
        sf_app_name = os.getenv("SF_TEST_APP_NAME")
        if sf_username is None or sf_password is None:
            raise ValueError(
                "SF_TEST_ACCOUNT_USERNAME, SF_TEST_ACCOUNT_PASSWORD, SF_TEST_ACCOUNT_NAME must be set if RAI_CLOUD_PROVIDER is set to 'snowflake'"
            )

        current_config = {
            "platform": "snowflake",
            "user": sf_username,
            "password": sf_password,
            "account": sf_account,
            "role": sf_role,
            "warehouse": sf_warehouse,
            "rai_app_name": sf_app_name,
            "use_package_manager": use_package_manager,
            "use_direct_access": use_direct_access,
        }
        if engine_name:
            current_config["engine"] = engine_name
        return cfg.Config(current_config, fetch=fetch_profile)

    else:
        raise ValueError(f"Unsupported cloud provider: {cloud_provider}")

def snowflake_handler_fixture():
    if not os.getenv("SF_REPORTING_PASSWORD"):
        print('snowflake logger disabled since required config env vars not present')
        yield
        return

    conn = snowflake.connector.connect(
        user=os.getenv("SF_REPORTING_USER"),
        password=os.getenv("SF_REPORTING_PASSWORD"),
        account=os.getenv("SF_REPORTING_ACCOUNT"),
        role=os.getenv("SF_REPORTING_ROLE"),
        warehouse=os.getenv("SF_REPORTING_WAREHOUSE"),
        database=os.getenv("SF_REPORTING_DATABASE"),
        schema=os.getenv("SF_REPORTING_SCHEMA"),
    )
    snowflake_handler = SnowflakeHandler(TRACE_ID, conn)
    logger.addHandler(snowflake_handler)
    yield
    snowflake_handler.shut_down()

def span_structure_validator_fixture():
    handler = SpanCollectorHandler()
    logger.addHandler(handler)
    yield handler
    logger.removeHandler(handler)
    # We skip tests in certain environments, so there's no need to perform span checks when those tests aren't executed.
    if os.getenv("SkipSpanValidation") != str(True):
        assert_valid_span_structure(handler.nodes, handler.events)

def root_span_fixture(get_full_config=False, span_type: str = "test_session", extra_attrs: Optional[dict] = None):
    root_span_attrs = get_root_span_attrs(get_full_config)
    if extra_attrs:
        root_span_attrs.update(extra_attrs)
    with debugging.span(span_type, **root_span_attrs):
        yield

def otel_collector_fixture():
    config = make_config()
    resources = rai.clients.snowflake.Resources(config=config)
    enable_otel_export(
        resources,
        config.get('rai_app_name', 'RELATIONALAI')
    )

    yield


from typing import Tuple
from relationalai.early_access.lqp import ir as lqp

def mk_and(args: list[lqp.Formula]) -> lqp.Formula:
    # Flatten nested conjunctions
    if any(isinstance(arg, lqp.Conjunction) for arg in args):
        final_args = []
        for arg in args:
            if isinstance(arg, lqp.Conjunction):
                final_args.extend(arg.args)
            else:
                final_args.append(arg)
        args = final_args

    if len(args) == 1:
        return args[0]

    return lqp.Conjunction(args=args, meta=None)

def mk_or(args: list[lqp.Formula]) -> lqp.Formula:
    # Flatten nested disjunctions
    if any(isinstance(arg, lqp.Disjunction) for arg in args):
        final_args = []
        for arg in args:
            if isinstance(arg, lqp.Disjunction):
                final_args.extend(arg.args)
            else:
                final_args.append(arg)
        args = final_args

    if len(args) == 1:
        return args[0]

    return lqp.Disjunction(args=args, meta=None)

def mk_abstraction(vars: list[Tuple[lqp.Var,lqp.RelType]], value: lqp.Formula) -> lqp.Abstraction:
    return lqp.Abstraction(vars=vars, value=value, meta=None)

def mk_exists(vars: list[Tuple[lqp.Var, lqp.RelType]], value: lqp.Formula) -> lqp.Formula:
    if len(vars) == 0:
        return value
    abstr = mk_abstraction(vars, value)
    return lqp.Exists(body=abstr, meta=None)

def mk_var(name: str) -> lqp.Var:
    return lqp.Var(name=name, meta=None)




import relationalai as rai
import string
import random
import textwrap

def rel(model, rel_code):
    resources = rai.clients.snowflake.Resources()
    resources.config.set("use_graph_index", False)
    database = model
    try:
        resources.create_graph(database)
    except Exception as e:
        if "already exists" not in str(e).lower():
            raise e
    engine = resources.config.get("engine")
    return resources.exec_raw(database, engine, rel_code)

def random_string():
    chars = string.ascii_letters + string.digits
    return ''.join(random.choice(chars) for _ in range(32))

def run_sql(session, query, params=None):
    return session.sql(query, params).collect()

def get_models(session, database, engine):
    APP_NAME = "relationalai"
    tmp_name = f"tmp_{random_string()}"
    query = textwrap.dedent(f"""
        call {APP_NAME}.api.exec_into(
            '{database}',
            '{engine}',
            'def pairs(name, model): rel(:catalog, :model, name, model) and not starts_with(name, "rel/") and not starts_with(name, "pkg/std")
             def Export_Relation(:key, i, key): exists( (value) | sort(pairs, i, key, value) )
             def Export_Relation(:value, i, value): exists( (key) | sort(pairs, i, key, value) )',
            '{tmp_name}',
            true
        );
    """)
    try:
        run_sql(session, query)
    except Exception as e:
        raise e
        return
    result = run_sql(session, f"select key, value from relationalai.results.{tmp_name};")
    run_sql(session, f"call {APP_NAME}.api.drop_result_table('{tmp_name}');")
    for row in result:
        if row["KEY"] == "catalog":
            continue
        yield row["KEY"], row["VALUE"]


import textwrap
from abc import abstractmethod
from collections import OrderedDict
from collections.abc import Iterable
from enum import Enum
from typing import Optional, Union

import relationalai.early_access.builder as qb
from relationalai.early_access.builder import define, where, annotations
from relationalai.early_access.dsl.bindings.common import Binding, BindableAttribute, IdentifierConceptBinding, \
    PartitioningConceptBinding, BindableTable, PrimitiveFilterBy, FilterBy, ReferentConceptBinding, Symbol
from relationalai.early_access.dsl.codegen.helpers import reference_entity, construct_entity
from relationalai.early_access.dsl.orm.constraints import Unique
from relationalai.early_access.dsl.orm.relationships import Role
from relationalai.early_access.rel.rel_utils import DECIMAL64_SCALE, DECIMAL128_SCALE


class GeneratedRelation(qb.Relationship):
    """
    A class representing a relation generated from a model.
    """
    def __init__(self, madlib, model, name):
        super().__init__(madlib, model=model.qb_model(), short_name=name)

    def __repr__(self):
        return f'GeneratedRelation({self._name})'


class InternallyGeneratedRelation(GeneratedRelation):
    """
    A class representing a relation generated by analyzing the model.
    """
    def __init__(self, madlib, dsl_model, name):
        super().__init__(madlib, dsl_model, name)


def filtering_view(row, filter_by: Optional[FilterBy], column_ref=None):
    """
    Generates filtering atoms for the map if the binding has any filters, which must be simple
    filters on the attribute view of the same table.
    """
    if not filter_by:
        return where()
    if isinstance(filter_by, PrimitiveFilterBy):
        assert column_ref is not None, 'Attribute must be provided if filter_by is a PrimitiveFilterBy'
        atoms = [_primitive_filtering_atom(column_ref, filter_by)]
    elif isinstance(filter_by, qb.Expression):
        atoms = [_expr_filtering_atom(row, filter_by)]
    elif isinstance(filter_by, Iterable):
        atoms = [_expr_filtering_atom(row, filter_expr) for filter_expr in filter_by]
    else:
        raise TypeError(f'Expected a PrimitiveFilterBy, Expression, or Iterable of Expressions, got {type(filter_by)}')
    return where(*atoms)

def _primitive_filtering_atom(column_ref, condition: PrimitiveFilterBy):
    """
    Generates a filtering atom for the map based on a primitive filter.
    """
    if isinstance(condition, Enum):
        condition = condition.name
    return where(
        column_ref == condition
    )

def _expr_filtering_atom(row, condition: qb.Expression):
    """
    Generates a filtering atom for the map based on a condition.
    """
    if not isinstance(condition, qb.Expression):
        raise TypeError(f'Expected an Expression, got {type(condition)}')
    params = condition._params
    if len(params) != 2:
        raise ValueError(f'Expected a condition with two parameters, got {len(params)}: {condition}')
    column, value = params
    if isinstance(value, Enum):
        value = value.name
    orig = column.type().ref()
    column_relation = Symbol(column.physical_name())
    return where(
        column(column_relation, row, orig),
        qb.Expression(condition._op, orig, value)
    )


class AbstractMap(GeneratedRelation):
    """
    A class representing an abstract map.
    """
    def __init__(self, madlib, dsl_model, name):
        super().__init__(madlib, dsl_model, name)

    @staticmethod
    def _filtering_atoms(conditions, row, column_ref=None):
        return filtering_view(row, conditions, column_ref)


class RoleMap(AbstractMap):
    """
    A class representing a value map relation.
    """

    def __init__(self, madlib, model, name, role, functional: bool = False):
        super().__init__(madlib, model, name)
        self._role = role

    def role(self):
        """
        Returns the role of the value map.
        """
        return self._role

    def value_player(self):
        """
        Returns the value player of the value map.
        """
        return self._role.player()

    @abstractmethod
    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this role map.
        """
        raise NotImplementedError("Subclasses must implement this method")

    def table(self) -> BindableTable:
        """
        Returns the table associated with the role map.
        """
        return self.column().table

    def __repr__(self):
        return f'RoleMap({self._name})'


class ValueMap(RoleMap):
    """
    A class representing a value map relation.
    """

    def __init__(self, model, binding: Binding, role, value_converter: Optional['ValueConverter']=None):
        madlib, name = self._handle_params(binding, role)
        super().__init__(madlib, model, name, role)
        self._binding = binding
        self._value_converter = value_converter
        self._generate_body()

    @staticmethod
    def _handle_params(binding, role):
        if not role.player()._is_primitive():
            raise TypeError(f'Cannot construct a value map for {role}, concept is not a value type')

        table_name = binding.column.table.physical_name()
        concept = role.player()
        name = f'{table_name}_row_to_{concept}'
        madlib = f'{name} {{row:Integer}} {{val:{concept}}}'
        return madlib, name

    def _generate_body(self):
        column = self._binding.column
        converter = self._value_converter
        row = qb.Integer.ref()
        result_type = self.value_player()
        if converter:
            val, converted = column.type().ref(), result_type.ref()
            filter_atoms = self._filtering_atoms(self._binding.filter_by, row, val)
            define(self(row, result_type(converted))).where(
                where(
                    column(row, val),
                    converter(val, converted),
                    filter_atoms
                )
            )
        else:
            val = column.type().ref()
            filter_atoms = self._filtering_atoms(self._binding.filter_by, row, val)
            define(
                self(row, result_type(val))
            ).where(
                column(row, val),
                filter_atoms
            )

    def column(self) -> BindableAttribute:
        return self._binding.column

    def __repr__(self):
        return f'ValueMap({self._name})'


class SimpleConstructorEntityMap(RoleMap):
    """
    A class representing an entity map relation.
    """

    def __init__(self, model, binding: Binding, role: Role, identifier_to_role_map: OrderedDict[Unique, 'RoleMap']):
        madlib, name = self._handle_params(binding, role)
        super().__init__(madlib, model, name, role)
        self._binding = binding
        self._identifier_to_role_map = identifier_to_role_map
        self._reference_role_map = list(identifier_to_role_map.items())[-1][1]  # last role map is the reference one
        self._generate_body()

    @staticmethod
    def _handle_params(binding, role):
        if role.player()._is_primitive():
            raise TypeError(f'Cannot construct an entity map for {role}, concept is not an entity type')

        table_name = binding.column.table.physical_name()
        concept = role.player()
        name = f'{table_name}_row_to_{concept}'
        madlib = f'{name} {{row:Integer}} {{val:{concept}}}'
        return madlib, name

    def _generate_body(self):
        concept = self.role().player()

        row = qb.Integer.ref()
        if not self._should_reference():
            # then create entities
            role_maps = self._identifier_to_role_map.values()
            values = [role_map.value_player().ref() for role_map in role_maps]
            where(
                *[role_map(row, value) for role_map, value in zip(role_maps, values)]
            ).define(construct_entity(concept, *values))

        # and populate the entity map, role_map is always the last one in the identifier_to_role_map
        value = self._reference_role_map.value_player().ref()
        where(
            self._reference_role_map(row, value), entity := reference_entity(concept, value),
            self._concept_population_atom(entity)
        ).define(self(row, entity))

    def _concept_population_atom(self, entity):
        """
        Generates the *optional* population atom for the entity map.
        Only used for bindings that have a FK.
        """
        if self._should_reference():
            return where(
                self.role().player()(entity)
            )
        else:
            return where()

    def _should_reference(self) -> bool:
        return self._reference_role_map.column().references_column is not None

    def binding(self) -> Binding:
        """
        Returns the binding of the entity map.
        """
        return self._binding

    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this entity map.
        """
        return self._binding.column

    def __repr__(self):
        return f'CtorEntityMap({self._name})'


class ReferentEntityMap(RoleMap):
    """
    A class representing a referent entity map relation.
    """

    def __init__(self, model, binding: Binding, role: Role, constructing_role_map: 'RoleMap'):
        madlib, name = self._handle_params(binding, role)
        super().__init__(madlib, model, name, role)
        self._binding = binding
        self._constructing_role_map = constructing_role_map
        self._generate_body()

    @staticmethod
    def _handle_params(binding, role):
        if role.player()._is_primitive():
            raise TypeError(f'Cannot construct an entity map for {role}, concept is not an entity type')

        table_name = binding.column.table.physical_name()
        concept = role.player()
        name = f'{table_name}_row_ref_to_{concept}'
        madlib = f'{name} {{row:Integer}} {{val:{concept}}}'
        return madlib, name

    def _generate_body(self):
        concept = self.role().player()

        row, value = qb.Integer.ref(), self._constructing_role_map.value_player().ref()
        where(
            self._constructing_role_map(row, value),
            entity := reference_entity(concept, value),
            concept(entity)
        ).define(self(row, entity))

    def binding(self) -> Binding:
        """
        Returns the binding of the entity map.
        """
        return self._binding

    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this entity map.
        """
        return self._binding.column

    def __repr__(self):
        return f'ReferentEntityMap({self._name})'

class CompositeEntityMap(AbstractMap):
    """
    A class representing a composite entity map relation.

    Takes a sequence of EntityMaps and constructs a composite entity type.
    """

    def __init__(self, model, entity_concept: qb.Concept, *role_maps: 'RoleMap'):
        madlib, name = self._handle_params(entity_concept, *role_maps)
        super().__init__(madlib, model, name)
        self._entity_concept = entity_concept
        self._role_maps = role_maps
        self._generate_body()

    def _handle_params(self, entity_concept: qb.Concept, *role_maps: 'RoleMap'):
        if entity_concept._is_primitive():
            raise TypeError(f'Cannot construct a composite entity map for {entity_concept},'
                            f' concept is not an entity type')
        if len(role_maps) < 2:
            raise ValueError('CompositeEntityMap requires at least two EntityMaps')

        role_map = role_maps[0]
        table = role_map.binding().column.table
        self._table = table
        name = f'{table.physical_name()}_row_to_{entity_concept}'
        madlib = f'{name} {{row:Integer}} {{val:{entity_concept}}}'
        return madlib, name

    def _generate_body(self):
        row = qb.Integer.ref()
        vars = [role_map.value_player().ref() for role_map in self._role_maps]

        # construct entities
        where(
            self._body_formula(row, *vars),
            entity := construct_entity(self._entity_concept, *vars),
        ).define(entity)

        # populate the entity map
        where(
            self._body_formula(row, *vars),
            entity := reference_entity(self._entity_concept, *vars),
            #
        ).define(self(row, entity))

    def _body_formula(self, row, *vars):
        return where(
            *[role_map(row, var) for role_map, var in zip(self._role_maps, vars)],
        )

    def value_player(self):
        """
        Returns the value player of the composite entity map.
        """
        return self._entity_concept

    def table(self):
        """
        Returns the table associated with the composite entity map.
        """
        return self._table

    def __repr__(self):
        return f'CompositeEntityMap({self._name})'


AbstractEntityMap = Union['SimpleConstructorEntityMap', 'ReferentEntityMap', 'CompositeEntityMap', 'UnionEntityMap']

class EntitySubtypeMap(AbstractMap):
    """
    A class representing an entity subtype map relation.
    """

    def __init__(self, model, binding: Union[IdentifierConceptBinding, PartitioningConceptBinding, ReferentConceptBinding],
                 ctor_entity_map: 'AbstractEntityMap'):
        madlib, name = self._handle_params(binding)
        super().__init__(madlib, model, name)
        self._binding = binding
        self._ctor_entity_map = ctor_entity_map
        self._generate_body()

    @staticmethod
    def _handle_params(binding: Union[IdentifierConceptBinding, PartitioningConceptBinding, ReferentConceptBinding]):
        table = binding.column.table
        entity_concept = binding.entity_type
        name = f'{table.physical_name()}_row_to_{entity_concept}'
        madlib = f'{name} {{row:Integer}} {{val:{entity_concept}}}'
        return madlib, name

    def _generate_body(self):
        subtype = self._binding.entity_type
        row, parent_entity = qb.Integer.ref(), self._ctor_entity_map.value_player().ref()

        # derive subtype population
        self._generate_body_formula(row, parent_entity).define(subtype(parent_entity))

        # populate the entity subtype map
        self._generate_body_formula(row, parent_entity).define(self(row, parent_entity))

    def _generate_body_formula(self, row, parent_entity):
        filtering_atom = self._get_filtering_atom(row)
        return where(
            self._ctor_entity_map(row, parent_entity),
            filtering_atom
        )

    def _get_filtering_atom(self, row):
        if isinstance(self._binding, PartitioningConceptBinding):
            filter_expr = self._binding.column == self._binding.has_value
            return self._filtering_atoms(filter_expr, row)
        else:
            return where()

    def _should_reference(self) -> bool:
        return self._role_map.column().references_column is not None

    def binding(self):
        """
        Returns the binding of the entity subtype map.
        """
        return self._binding

    def value_player(self):
        """
        Returns the subtype of the entity subtype map.
        """
        return self._binding.entity_type

    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this entity subtype map.
        """
        return self._binding.column

    def table(self) -> BindableTable:
        """
        Returns the table associated with the entity subtype map.
        """
        return self._binding.column.table

    def __repr__(self):
        return f'EntitySubtypeMap({self._name})'


class UnionEntityMap(AbstractMap):
    """
    A class representing a union entity map relation.
    """

    def __init__(self, model, entity_concept: qb.Concept, *entity_maps: AbstractEntityMap, generate_population: bool=False):
        madlib, name = self._handle_params(entity_concept, *entity_maps)
        super().__init__(madlib, model, name)
        self._entity_type = entity_concept
        self._entity_maps = list(entity_maps)
        self._generate_population = generate_population
        self._generate_body()

    def _handle_params(self, entity_concept: qb.Concept, *entity_maps: AbstractEntityMap):
        if len(entity_maps) == 0:
            raise ValueError('UnionEntityMap requires at least one EntityMap')
        # pick an arbitrary entity map to get the table, as they must all have the same
        table = entity_maps[0].table()
        self._table = table
        name = f'{table.physical_name()}_row_to_{entity_concept}'
        madlib = f'{name} {{row:Integer}} {{val:{entity_concept}}}'
        return madlib, name

    def _generate_body(self):
        for entity_map in self._entity_maps:
            self._generate_body_rule(entity_map)

    def _generate_body_rule(self, entity_map: AbstractEntityMap):
        row, entity = qb.Integer.ref(), entity_map.value_player().ref()

        if self._generate_population:
            # derive type population
            where(
                entity_map(row, entity)
            ).define(self._entity_type(entity))

        # derive union entity map
        where(
            entity_map(row, entity)
        ).define(self(row, entity))

    def value_player(self):
        """
        Returns the type of the entity map.
        """
        return self._entity_type

    def table(self):
        """
        Returns the table associated with the entity map.
        """
        return self._table

    def update(self, entity_map: AbstractEntityMap):
        """
        Updates the union entity map with a new entity map.
        """
        if entity_map._id in [existing._id for existing in self._entity_maps]:
            return
        self._entity_maps.append(entity_map)
        self._generate_body_rule(entity_map)

    def __repr__(self):
        return f'UnionEntityMap({self._name})'


class ValueConverter(InternallyGeneratedRelation):
    """
    Base class for value converter relations.
    """
    def __init__(self, madlib, dsl_model, name):
        super().__init__(madlib, dsl_model, name)
        self.annotate(annotations.external)

    @abstractmethod
    def result_type(self) -> qb.Concept:
        pass

# Note: this conversion is only needed until CDC standardizes on the scale and size of the Decimal type.

class DecimalValueConverter(ValueConverter):
    """
    A class representing a decimal value converter relation.
    """

    def __init__(self, model, size_from: int, scale_from: int, type_to: qb.Concept):
        madlib, name = self._generate_madlib_and_name(size_from, scale_from, type_to)
        super().__init__(madlib, model, name)
        self._validate(type_to)
        self._size_from = size_from
        self._scale_from = scale_from
        self._type_to = type_to
        self._size_to = 64 if type_to is qb.Decimal64 else 128
        self._scale_to = DECIMAL64_SCALE if type_to is qb.Decimal64 else DECIMAL128_SCALE
        self._generate_body()

    @staticmethod
    def _validate(type_to):
        if type_to is not qb.Decimal64 and type_to is not qb.Decimal128:
            raise TypeError(f'Expected Decimal64 or Decimal128, got {type_to}')

    def result_type(self) -> qb.Concept:
        return self._type_to

    @staticmethod
    def _generate_madlib_and_name(size_from: int, scale_from: int, type_to: qb.Concept):
        name = f'value_converter_{type_to}_from_{size_from}_{scale_from}'
        madlib = f'{name} {{Decimal}} {{{type_to}}}'
        return madlib, name

    def _generate_body(self):
        src = textwrap.dedent(f"""
        @inline
        def {self._name}(orig, rez):
            exists((t_size, t_scale, size, scale, int_val) |
                ^FixedDecimal(t_size, t_scale, int_val, orig) and
                ::std::mirror::lower(t_size, size) and
                ::std::mirror::lower(t_scale, scale) and
                decimal({self._size_to}, {self._scale_to}, int_val / (10 ^ scale), rez)
            )""")
        assert self._model is not None, "Model must be defined before defining a relation"
        self._model.define(qb.RawSource('rel', src))

    def __repr__(self):
        return f'DecimalValueConverter({self._name})'



from __future__ import annotations
from collections import defaultdict
from dataclasses import dataclass
import dataclasses
import datetime
from decimal import Decimal as PyDecimal
from typing import Optional, Union
from relationalai import debugging
from relationalai.early_access.metamodel import builtins, helpers, ir, types, visitor, compiler, factory as f
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
import rich

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def to_relation(node:ir.Lookup|ir.Update|ir.Aggregate) -> ir.Relation:
    if isinstance(node, ir.Aggregate):
        return node.aggregation
    return node.relation

def to_name(type:ir.Type) -> str:
    if isinstance(type, ir.ScalarType):
        return type.name
    elif isinstance(type, ir.UnionType):
        return '|'.join([to_name(t) for t in type.types])
    else:
        raise TypeError(f"Unknown type: {type}")

#--------------------------------------------------
# Constants
#--------------------------------------------------

# map literal types to IR types
TYPE_MAPPING = {
    str: types.String,
    bool: types.Bool,
    int: types.Int,
    float: types.Float,
    PyDecimal: types.Decimal128,
    datetime.date: types.Date,
    datetime.datetime: types.DateTime,
}

# map potential type conversions to their cost
CONVERSION_LATTICE:dict[tuple[ir.Type|None, ir.Type|None], float] = {
    # Int -> Decimal64 -> Decimal128 -> Float
    (types.Int, types.Decimal64): 3.0,
    (types.Decimal64, types.Decimal128): 5.0,
    (types.Decimal128, types.Float): 7.0,
    # transitive closure of the above
    (types.Int, types.Decimal128): 8.0,
    (types.Decimal64, types.Float): 12.0,
    (types.Int, types.Float): 15.0,
}

PRIMITIVES = [types.Int, types.Float, types.Decimal64, types.Decimal128, types.Bool, types.String, types.Date, types.DateTime]

ENTITY_LIKES = [types.RowId]

# these are relations that try to preserve their input types
# e.g. if you add a USD and a Decimal, the result is still a USD
TYPE_PRESERVERS = [
    *builtins.plus.overloads,
    *builtins.minus.overloads,
    *builtins.mul.overloads,
    *builtins.mod.overloads,
    builtins.trunc_div,
    builtins.concat,
    *builtins.abs.overloads,
    *builtins.sum.overloads,
    *builtins.max.overloads,
    *builtins.min.overloads,
    # Exclude div overloads (like Int/Int -> Float) where the result type is different from the input types
    *(filter(lambda x: x.fields[0].type == x.fields[1].type == x.fields[2].type, builtins.div.overloads)),
    # Exclude avg overloads (avg Int -> Float) where the result type is different from the input types
    *(filter(lambda x: x.fields[0].type == x.fields[1].type, builtins.avg.overloads)),
]

#--------------------------------------------------
# Type helpers
#--------------------------------------------------

def is_abstract(type:ir.Type) -> bool:
    return types.is_abstract_type(type)

def to_type(value: ir.Value|ir.Field) -> ir.Type:
    if isinstance(value, (ir.Var, ir.Field, ir.Literal)):
        return value.type

    value_type = type(value)
    if value_type in TYPE_MAPPING:
        return TYPE_MAPPING[value_type]

    if isinstance(value, tuple):
        return types.AnyList

    raise TypeError(f"Cannot determine IR type for value: {value} of type {type(value).__name__}")

def type_matches(actual:ir.Type, expected:ir.Type, allow_expected_parents=False) -> bool:
    is_scalar = isinstance(actual, ir.ScalarType)
    exp_is_scalar = isinstance(expected, ir.ScalarType)
    if actual == expected:
        return True
    if is_scalar and any([type_matches(parent, expected) for parent in actual.super_types]):
        return True
    if allow_expected_parents and exp_is_scalar and any([type_matches(actual, parent, allow_expected_parents) for parent in expected.super_types]):
        return True
    if not is_primitive(expected) and exp_is_scalar and actual in ENTITY_LIKES:
        return True
    if isinstance(actual, ir.UnionType):
        for t in actual.types:
            if type_matches(t, expected):
                return True
    if actual == types.Any or expected == types.Any:
        return True
    if actual == types.Number and expected in (types.Int, types.Float, types.Decimal64, types.Decimal128):
        return True
    if expected == types.Number and actual in (types.Int, types.Float, types.Decimal64, types.Decimal128):
        return True
    return False

def conversion_allowed(arg:ir.Value, actual:ir.Type, expected:ir.Type, arg_types:set[ir.Type]=set()) -> bool:
    return conversion_cost(arg, actual, expected, arg_types) < float("inf")

def conversion_cost(arg: ir.Value, actual:ir.Type, expected:ir.Type, arg_types:set[ir.Type]) -> float:
    if type_matches(actual, expected):
        return 0

    # if we have a type variable and all the types in the expression
    # match, then there's no promotion needed and this is a valid overload
    elif expected is types.EntityTypeVar and not is_primitive(actual) and len(arg_types) == 1:
        return 0

    # Literal numbers can be converted to other number types.
    if isinstance(arg, ir.Literal) or isinstance(arg, (int, float, PyDecimal)):
        if actual == types.Int and expected in (types.Float, types.Decimal64, types.Decimal128):
            return 1.0
        elif actual == types.Float and expected in (types.Decimal64, types.Decimal128):
            return 1.0
        elif actual in (types.Decimal64, types.Decimal128) and expected in (types.Decimal64, types.Decimal128):
            return 1.0

    base_actual = to_base_primitive(actual)
    base_expected = to_base_primitive(expected)
    if (base_actual, base_expected) in CONVERSION_LATTICE:
        return CONVERSION_LATTICE[(base_actual, base_expected)]
    return float("inf")

def merge_types(type1: ir.Type, type2: ir.Type) -> ir.Type:
    combined = ordered_set()
    types_to_process = [type1, type2]

    # Iterative flattening of union types
    while types_to_process:
        t = types_to_process.pop()
        if isinstance(t, ir.UnionType):
            types_to_process.extend(t.types)
        else:
            combined.add(t)

    # If we have multiple types and Any is one of them, remove Any
    if len(combined) > 1 and types.Any in combined:
        combined.remove(types.Any)

    # Return single type or create a union
    return next(iter(combined)) if len(combined) == 1 else f.union_type(list(combined))

def to_base_primitive(type:ir.Type) -> Optional[ir.Type]:
    if isinstance(type, ir.ScalarType):
        if is_base_primitive(type):
            return type
        # walk the hierarchy to find a base primitive
        for parent in type.super_types:
            if found := to_base_primitive(parent):
                return found
    if isinstance(type, ir.UnionType):
        for t in type.types:
            if found := to_base_primitive(t):
                return found
    return None

def is_base_primitive(type:ir.Type) -> bool:
    return type in PRIMITIVES

def is_primitive(type:ir.Type) -> bool:
    return to_base_primitive(type) is not None

def invalid_type(type:ir.Type) -> bool:
    if isinstance(type, ir.UnionType):
        # if there are multiple primitives, or a primtive and a non-primitive
        # then we have an invalid type
        if len(type.types) > 1:
            return any([is_primitive(t) for t in type.types])
    return False

def try_preserve_type(types:set[ir.Type]) -> Optional[ir.Type]:
    # we keep the input type as the output type if either all inputs
    # are the exact same type or there's one nominal and its base primitive
    # type, e.g. USD + Decimal
    if len(types) == 1:
        return next(iter(types))
    if len(types) == 2:
        t1, t2 = types
        base_equivalent = type_matches(t1, t2, allow_expected_parents=True)
        if base_equivalent:
            # as long as one of the types is a base primitive, we can use the
            # other type as final preserved type
            if is_base_primitive(t1):
                return t2
            elif is_base_primitive(t2):
                return t1
    return None

#--------------------------------------------------
# Type Errors
#--------------------------------------------------

@dataclass
class TyperError():
    node: ir.Node

@dataclass
class TypeMismatch(TyperError):
    expected: ir.Type
    actual: ir.Type
    def __str__(self):
        return f"[red bold][Type Mismatch][/red bold] expected [yellow]{to_name(self.expected)}[/yellow], got [red]{to_name(self.actual)}[/red]: [white]{str(self.node).strip()}[/white]"

@dataclass
class InvalidType(TyperError):
    type: ir.Type
    def __str__(self):
        return f"[red bold][Invalid Type][/red bold] incompatible types [red]{to_name(self.type)}[/red]: [white]{str(self.node).strip()}[/white]"

@dataclass
class UnresolvedOverload(TyperError):
    arg_types: list[ir.Type]
    def __str__(self):
        assert isinstance(self.node, (ir.Lookup, ir.Update, ir.Aggregate))
        rel = to_relation(self.node)
        types = ', '.join([to_name(t) for t in self.arg_types])
        return f"[red bold][Unresolved Overload][/red bold] [yellow]{rel.name}({types})[/yellow]: [white]{str(self.node).strip()}[/white]"

@dataclass
class UnresolvedType(TyperError):
    def __str__(self):
        return f"[red bold][Unresolved Type][/red bold] [white]{str(self.node).strip()}[/white]"
    pass

#--------------------------------------------------
# Propagation Network
#--------------------------------------------------

# The core idea of the typer is to build a propagation network where nodes
# are vars, fields, or overloaded lookups/updates/aggregates. The intuition
# is that _all_ types in the IR ultimately flow from relation fields, so if
# we figure those out we just need to propagate their types to unknown vars, which
# may then flow into other fields and so on.

# This means the network only needs to contain nodes that either directly flow into
# an abstract node or are themselves abstract. We need to track overloads because
# their arguments effectively act like abstract vars until we've resolved the final types.

Node = Union[ir.Var, ir.Field, ir.Lookup, ir.Update, ir.Aggregate]

class PropagationNetwork():
    def __init__(self):

        # track the set of nodes that represent entry points into the network
        self.roots = ordered_set()
        # we separately want to track nodes that were loaded from a previous run
        # so that even if we have edges to them, we _still_ consider them roots
        # and properly propagate types from them at the beginning
        self.loaded_roots = set()
        self.edges:dict[Node, OrderedSet[Node]] = defaultdict(lambda: ordered_set())
        self.has_incoming = set()
        self.type_requirements:dict[Node, OrderedSet[ir.Field]] = defaultdict(lambda: ordered_set())

        self.errors:list[TyperError] = []
        self.resolved_types:dict[Node, ir.Type] = {}
        self.resolved_overloads:dict[int, list[ir.Relation]] = {}

    #--------------------------------------------------
    # Mismatches
    #--------------------------------------------------

    def mismatch(self, node:Node, expected:ir.Type, actual:ir.Type):
        self.errors.append(TypeMismatch(node, expected, actual))

    def invalid_type(self, node:Node, type:ir.Type):
        self.errors.append(InvalidType(node, type))

    def unresolved_overload(self, node:ir.Lookup|ir.Update|ir.Aggregate):
        self.errors.append(UnresolvedOverload(node, [self.resolve(a) for a in node.args]))

    def unresolved_type(self, node:Node):
        self.errors.append(UnresolvedType(node))

    def has_errors(self, node:Node) -> bool:
        for mismatch in self.errors:
            if mismatch.node == node:
                return True
        return False

    #--------------------------------------------------
    # Types and Edges
    #--------------------------------------------------

    def add_edge(self, source:Node, target:Node):
        if target not in self.loaded_roots:
            self.roots.remove(target)
        if source not in self.has_incoming:
            self.roots.add(source)
        self.edges[source].add(target)
        self.has_incoming.add(target)

    def add_type(self, node:Node, type:ir.Type):
        if node in self.resolved_types:
            self.resolved_types[node] = merge_types(self.resolved_types[node], type)
        else:
            self.resolved_types[node] = type

    def add_type_requirement(self, source:Node, field:ir.Field):
        self.type_requirements[source].add(field)

    #--------------------------------------------------
    # Load previous types
    #--------------------------------------------------

    def load_types(self, type_dict:dict[Node, ir.Type]):
        for node, type in type_dict.items():
            if isinstance(node, (ir.Field)):
                self.add_type(node, type)
                self.loaded_roots.add(node)
                self.roots.add(node)

    #--------------------------------------------------
    # Resolve
    #--------------------------------------------------

    def resolve(self, value:Node|ir.Value) -> ir.Type:
        if isinstance(value, (ir.Var, ir.Field)):
            return self.resolved_types.get(value) or to_type(value)
        assert not isinstance(value, (ir.Lookup, ir.Update, ir.Aggregate)), "Should never try to resolve a task"
        return to_type(value)

    #--------------------------------------------------
    # Overloads
    #--------------------------------------------------

    def resolve_overload_deps(self, op:ir.Lookup|ir.Update|ir.Aggregate, keep_known=False):
        # We need to find which args flow into this task, which flow out, and
        # if some are inputs, then they are required before we can resolve
        incoming:list[Node] = []
        outgoing:list[Node] = []
        required:list[Node] = []
        relation = to_relation(op)
        field_args = zip(relation.fields, [self.resolve(a) for a in op.args], op.args)
        has_inputs = any([field.input for field in relation.fields])
        for field, arg_type, arg in field_args:
            if (is_abstract(arg_type) or keep_known) and isinstance(arg, ir.Var):
                if field.input:
                    incoming.append(arg)
                    required.append(arg)
                elif not has_inputs:
                    incoming.append(arg)
                    outgoing.append(arg)
                else:
                    outgoing.append(arg)

        # check if the overloads of this relation are fully resolved, if any aren't
        # then their fields are incoming edges to this node
        for overload in relation.overloads:
            for field in overload.fields:
                if is_abstract(self.resolve(field)):
                    incoming.append(field)
                    required.append(field)

        return incoming, outgoing, required

    def resolve_overload(self, op:ir.Lookup|ir.Update|ir.Aggregate) -> Optional[list[ir.Relation]]:
        # check if we have any unresolved args that are required, if all of our args
        # are unresolved (len(incoming) == len(relation.fields)) then we have no information
        # to try and resolve the overloads with
        incoming, outgoing, required = self.resolve_overload_deps(op)
        relation = to_relation(op)
        if required or len(incoming) == len(relation.fields):
            return

        overloads = relation.overloads
        if not overloads:
            overloads = [relation]

        # otherwise we compute the cost of each overload and return the set of relations
        # that have the lowest cost. This can be multiple in e.g. the Person.pets.name
        # case where Cat_name and Dog_name are equally valid
        inf = float("inf")
        min_cost = inf
        matches = []
        for overload in overloads:
            arg_types = set(self.resolve(arg) for arg, field in zip(op.args, overload.fields) if field.input)
            total = 0
            for arg, field in zip(op.args, overload.fields):
                arg_type = self.resolve(arg)
                field_type = self.resolve(field)
                total += conversion_cost(arg, arg_type, field_type, arg_types)
            if total != inf and total <= min_cost:
                if total < min_cost:
                    min_cost = total
                    matches.clear()
                matches.append(overload)

        return matches

    #--------------------------------------------------
    # Propagation
    #--------------------------------------------------

    def propagate(self):
        edges = self.edges
        work_list = []

        # go through all the roots and find any that are not abstract, they'll
        # be the first nodes to push types through the network
        unhandled = ordered_set()
        for node in self.roots:
            if not isinstance(node, (ir.Var, ir.Field)):
                continue
            node_type = self.resolve(node)
            if not is_abstract(node_type):
                work_list.append(node)
            else:
                unhandled.add(node)

        # push known type nodes through the edges
        for source in work_list:
            unhandled.remove(source)
            source_type = self.resolve(source)
            # check to see if the source has ended up with a set of types that
            # aren't valid, e.g. a union of primitives
            if invalid_type(source_type):
                self.invalid_type(source, source_type)

            # propagate our type to each outgoing edge
            for out in edges.get(source, []):
                # if this is an overload then we need to try and resolve it
                if isinstance(out, (ir.Update, ir.Lookup, ir.Aggregate)):
                    found = self.resolve_overload(out)
                    if found is not None:
                        start_arg_types = [self.resolve(arg) for arg in out.args]
                        self.propagate_overloads(out, found)
                        final_arg_types = [self.resolve(arg) for arg in out.args]
                        for arg, start, final in zip(out.args, start_arg_types, final_arg_types):
                            if start != final:
                                work_list.append(arg)
                # otherwise, we just add to the outgoing node's type and if it
                # changes we add it to the work list
                elif start := self.resolve(out):
                    self.add_type(out, source_type)
                    if start != self.resolve(out) or out in unhandled:
                        work_list.append(out)

        for source in unhandled:
            self.unresolved_type(source)

        # now that we've pushed all the types through the network, we need to validate
        # that all type requirements of those nodes are met
        for node, fields in self.type_requirements.items():
            node_type = self.resolve(node)
            for field in fields:
                field_type = self.resolve(field)
                if not type_matches(node_type, field_type):
                    node_base = to_base_primitive(node_type)
                    field_base = to_base_primitive(field_type)
                    if (node_base, field_base) in CONVERSION_LATTICE:
                        pass
                    else:
                        self.mismatch(node, field_type, node_type)

        return self.resolved_types


    def propagate_overloads(self, node:ir.Lookup|ir.Update|ir.Aggregate, overloads:list[ir.Relation]):
        if not overloads:
            return self.unresolved_overload(node)

        # we've resolved the overloads, so store that
        self.resolved_overloads[node.id] = overloads

        # we need to determine the final types of our args but taking all the overloads
        # and adding the type of their fields back to the args.
        resolved_args = [self.resolve(a) for a in node.args]
        for overload in overloads:
            resolved_fields = [self.resolve(f) for f in overload.fields]
            # if our overload preserves types, we check to see if there's a preserved
            # output type given the inputs and if so, shadow the field's type with the
            # preserved type
            if overload in TYPE_PRESERVERS:
                input_types = set([arg_type for field, arg_type
                                    in zip(overload.fields, resolved_args)
                                    if field.input])
                if out_type := try_preserve_type(input_types):
                    resolved_fields = [field_type if field.input else out_type
                                        for field, field_type in zip(overload.fields, resolved_fields)]
            for field_type, arg_type, arg in zip(resolved_fields, resolved_args, node.args):
                if is_abstract(arg_type) and isinstance(arg, ir.Var):
                    self.add_type(arg, field_type)



    #--------------------------------------------------
    # Display
    #--------------------------------------------------

    # draw the network as a mermaid graph for the debugger
    def to_mermaid(self, max_edges=500) -> str:
        resolved = self.resolved_types
        nodes = ordered_set()
        link_strs = []
        for src, dsts in self.edges.items():
            nodes.add(src)
            for dst in dsts:
                if len(link_strs) > max_edges:
                    break
                nodes.add(dst)
                link_strs.append(f"n{src.id} --> n{dst.id}")
            if len(link_strs) > max_edges:
                break

        for src, dsts in self.type_requirements.items():
            nodes.add(src)
            for dst in dsts:
                if len(link_strs) > max_edges:
                    break
                nodes.add(dst)
                link_strs.append(f"n{src.id} --> n{dst.id}")
            if len(link_strs) > max_edges:
                break

        def type_span(t:ir.Type) -> str:
            type_str = t.name if isinstance(t, ir.ScalarType) else str(t)
            return f"<span style='color:cyan;'>{type_str.strip()}</span>"

        def overload_span(rel:ir.Relation, arg_types:list[ir.Type]) -> str:
            args = []
            for field, arg_type in zip(rel.fields, arg_types):
                field_type = self.resolve(field)
                if not type_matches(arg_type, field_type) and field_type != types.EntityTypeVar:
                    args.append(f"<span style='color:yellow;'>{str(arg_type).strip()} -> {str(field_type).strip()}</span>")
                elif isinstance(arg_type, ir.UnionType):
                    args.append(type_span(field_type))
                else:
                    args.append(type_span(arg_type))
            return f'{rel.name}({", ".join(args)})'

        node_strs = []
        for node in nodes:
            klass = ""
            if isinstance(node, ir.Var):
                ir_type = resolved.get(node) or self.resolve(node)
                type_str = type_span(ir_type)
                # if this is one of our vars that is a placeholder for a literal, just use the type
                if isinstance(ir_type, ir.ScalarType) and node.name == ir_type.name:
                    label = f'(["Literal {type_str}"])'
                else:
                    label = f'(["{node.name}: {type_str}"])'
            elif isinstance(node, ir.Field):
                ir_type = resolved.get(node) or self.resolve(node)
                type_str = type_span(ir_type)
                klass = ":::field"
                label = f"{{{{\"{node.name}: {type_str}\"}}}}"
            elif isinstance(node, (ir.Lookup, ir.Update, ir.Aggregate)):
                arg_types = [self.resolve(arg) for arg in node.args]
                if node.id in self.resolved_overloads:
                    overloads = self.resolved_overloads[node.id]
                    content = "<br/>".join([overload_span(o, arg_types) for o in overloads])
                else:
                    content = overload_span(to_relation(node), arg_types)
                label = f'[/"{content}"/]'
            else:
                raise NotImplementedError(f"Unknown node type: {type(node)}")
            if self.has_errors(node):
                klass = ":::error"
            node_strs.append(f"n{node.id}{label}{klass}")

        node_str = "\n                ".join(node_strs)
        link_str = "\n                ".join(link_strs)
        template = f"""
            %%{{init: {{'theme':'dark', 'flowchart':{{'useMaxWidth':false, 'htmlLabels': true}}}}}}%%
            flowchart TD
                linkStyle default stroke:#666
                classDef field fill:#245,stroke:#478
                classDef error fill:#624,stroke:#945,color:#f9a
                classDef default stroke:#444,stroke-width:2px, font-size:12px

                %% nodes
                {node_str}

                %% edges
                {link_str}
        """
        return template

    # simplified, less verbose (compared to mermaid) output for snapshot testing
    def to_fish(self) -> str:
        resolved = self.resolved_types
        nodes = ordered_set()
        for src, dsts in self.edges.items():
            nodes.add(src)
            nodes.update(dsts)
        for src, dsts in self.type_requirements.items():
            nodes.add(src)
            nodes.update(dsts)

        def type_to_str(t:ir.Type) -> str:
            type_str = t.name if isinstance(t, ir.ScalarType) else str(t)
            return type_str.strip()

        def overload_str(rel:ir.Relation, arg_types:list[ir.Type]) -> str:
            args = []
            for field, arg_type in zip(rel.fields, arg_types):
                field_type = self.resolve(field)
                if not type_matches(arg_type, field_type):
                    args.append(f"{str(arg_type).strip()} -?> {str(field_type).strip()}")
                elif isinstance(arg_type, ir.UnionType):
                    args.append(type_to_str(field_type))
                else:
                    args.append(type_to_str(arg_type))
            return f'{rel.name}({", ".join(args)})'

        def node_kind(node:ir.Node) -> str:
            if isinstance(node, (ir.Lookup, ir.Update, ir.Aggregate)):
                return "overload"
            return type(node).__name__.lower()

        def error_info(err:TyperError) -> str:
            if isinstance(err, TypeMismatch):
                return f'Type Mismatch|expected {to_name(err.expected)}, got {to_name(err.actual)}'
            return f'{type(err).__name__}'

        nl = "\n"
        node_strs = []
        for node in nodes:
            info = ""
            if isinstance(node, ir.Var):
                ir_type = resolved.get(node) or self.resolve(node)
                if not(isinstance(ir_type, ir.ScalarType) and node.name == ir_type.name):
                    info = f'{node.name}|{type_to_str(ir_type)}'
            elif isinstance(node, ir.Field):
                ir_type = resolved.get(node) or self.resolve(node)
                info = f'{node.name}|{type_to_str(ir_type)}'
            elif isinstance(node, (ir.Lookup, ir.Update, ir.Aggregate)):
                arg_types = [self.resolve(arg) for arg in node.args]
                if node.id in self.resolved_overloads:
                    overloads = self.resolved_overloads[node.id]
                    content = nl.join([overload_str(o, arg_types) for o in overloads])
                else:
                    content = overload_str(to_relation(node), arg_types)
                info = f'{content}'
            else:
                raise NotImplementedError(f"Unknown node type: {type(node)}")

            if info:
                error_suffix = " !" if self.has_errors(node) else ""
                node_strs.append(f'{node_kind(node)}|{info}{error_suffix}')

        node_strs.sort()
        if self.errors:
            node_strs.append("---ERRORS---")
            for err in self.errors:
                node_strs.append(f'{error_info(err)} ({str(err.node).strip()})')

        return f"""{nl.join(node_strs)}"""

#--------------------------------------------------
# Analyzer
#--------------------------------------------------

class Analyzer(visitor.Visitor):
    def __init__(self):
        super().__init__()
        self.net = PropagationNetwork()

        # this is a map of literal types to a var representing that literal in
        # the graph. This allows us to collapse all literals into a single node
        # and avoids exploding node count if lots of data gets added directly
        self.shared_literal_vars:dict[ir.Type, ir.Var] = {
            t: f.var(t.name, t)
            for t in types.builtin_types
            if isinstance(t, ir.ScalarType)
        }

    #--------------------------------------------------
    # Literal types
    #--------------------------------------------------

    # given a literal type, lookup or create a var for it
    def shared_literal_var(self, node:ir.Type) -> ir.Var:
        assert isinstance(node, ir.ScalarType)
        if node not in self.shared_literal_vars:
            self.shared_literal_vars[node] = f.var(node.name, node)
        return self.shared_literal_vars[node]

    #--------------------------------------------------
    # Lookups + Aggregates
    #--------------------------------------------------

    def visit_lookup(self, node: ir.Lookup, parent: Optional[ir.Node]):
        self.visit_rel_op(node, parent)

    def visit_aggregate(self, node: ir.Aggregate, parent: ir.Node | None):
        self.visit_rel_op(node, parent)

    def visit_rel_op(self, node: ir.Lookup|ir.Aggregate, parent: Optional[ir.Node]):
        rel = to_relation(node)
        if isinstance(node, ir.Lookup) and rel == builtins.eq:
            return self.visit_eq(node, parent)

        if rel.overloads:
            return self.visit_overloaded(node, parent)

        # if this is a population check, then it's fine to pass a subtype in to do the check
        # e.g. Employee(Person) is a valid way to check if a person is an employee
        allow_any_parents = builtins.concept_relation_annotation in rel.annotations
        arg_types = set(self.net.resolve(arg) for arg, field in zip(node.args, rel.fields) if field.input)
        for arg, field in zip(node.args, rel.fields):
            field_type = field.type
            arg_type = self.net.resolve(arg)
            if not type_matches(arg_type, field_type, allow_any_parents):
                # Do not complain if we can convert the arg to the field type.
                if not conversion_allowed(arg, arg_type, field_type, arg_types):
                    self.net.mismatch(node, field_type, arg_type)
            # if we have an abstract var then this field will ultimately propagate to that
            # var's type
            elif isinstance(arg, ir.Var) and is_abstract(arg_type) and not field.input:
                self.net.add_edge(field, arg)

    #--------------------------------------------------
    # Overloads
    #--------------------------------------------------

    def visit_overloaded(self, node: ir.Lookup|ir.Update|ir.Aggregate, parent: Optional[ir.Node]):
        incoming, outgoing, required = self.net.resolve_overload_deps(node, keep_known=True)
        for arg in incoming:
            self.net.add_edge(arg, node)
        for arg in outgoing:
            self.net.add_edge(node, arg)

    #--------------------------------------------------
    # Eq
    #--------------------------------------------------

    def visit_eq(self, node: ir.Lookup, parent: Optional[ir.Node]):
        (left, right) = node.args
        left_type = self.net.resolve(left)
        right_type = self.net.resolve(right)
        if is_abstract(left_type) and is_abstract(right_type):
            assert isinstance(left, ir.Var) and isinstance(right, ir.Var)
            # if both sides are abstract, then whatever we find out about
            # either should propagate to the other
            self.net.add_edge(left, right)
            self.net.add_edge(right, left)
        elif is_abstract(left_type):
            assert isinstance(left, ir.Var)
            if isinstance(right, ir.Var):
                self.net.add_edge(right, left)
            else:
                literal_var = self.shared_literal_var(right_type)
                self.net.add_edge(literal_var, left)
        elif is_abstract(right_type):
            assert isinstance(right, ir.Var)
            if isinstance(left, ir.Var):
                self.net.add_edge(left, right)
            else:
                literal_var = self.shared_literal_var(left_type)
                self.net.add_edge(literal_var, right)
        elif not type_matches(left_type, right_type):
            if not conversion_allowed(left, left_type, right_type) and not conversion_allowed(right, right_type, left_type):
                self.net.mismatch(node, left_type, right_type)

    #--------------------------------------------------
    # Update
    #--------------------------------------------------

    def visit_update(self, node: ir.Update, parent: Optional[ir.Node]):
        rel = node.relation
        # if this is a population check, then it's fine to pass a subtype in to do the population
        # e.g. Employee(Person) should be a valid way to populate a person
        allow_any_parents = builtins.concept_relation_annotation in rel.annotations
        for arg, field in zip(node.args, rel.fields):
            field_type = field.type
            arg_type = self.net.resolve(arg)
            is_var = isinstance(arg, ir.Var)

            # if the arg is abstract, but the field isn't, then we need to make
            # sure that once the arg is resolved we check that it matches the field
            # type
            if is_var and is_abstract(arg_type) and not is_abstract(field_type):
                self.net.add_type_requirement(arg, field)

            # if the field is abstract, then eventually this var will help determine
            # the field's type
            if is_abstract(field_type):
                source = self.shared_literal_var(arg_type) if not is_var else arg
                assert source
                self.net.add_edge(source, field)

            elif not type_matches(arg_type, field_type, allow_expected_parents=allow_any_parents):
                if not conversion_allowed(arg, arg_type, field_type):
                    self.net.mismatch(node, field_type, arg_type)

#--------------------------------------------------
# Replacer
#--------------------------------------------------

CONVERSIONS:dict[tuple[ir.Type|None, ir.Type|None], ir.Relation] = {
    (types.Int, types.Float): builtins.int_to_float,
    (types.Int, types.Decimal64): builtins.int_to_decimal64,
    (types.Int, types.Decimal128): builtins.int_to_decimal128,
    (types.Float, types.Int): builtins.float_to_int,
    (types.Float, types.Decimal64): builtins.float_to_decimal64,
    (types.Float, types.Decimal128): builtins.float_to_decimal128,
    (types.Decimal64, types.Int): builtins.decimal64_to_int,
    (types.Decimal64, types.Float): builtins.decimal64_to_float,
    (types.Decimal64, types.Decimal128): builtins.decimal64_to_decimal128,
    (types.Decimal128, types.Int): builtins.decimal128_to_int,
    (types.Decimal128, types.Float): builtins.decimal128_to_float,
    (types.Decimal128, types.Decimal64): builtins.decimal128_to_decimal64,
}

# Once we've pushed all the types through the network, we need to
# to replace the types of fields and vars that we may have discovered.
# We also need to replace overloaded lookups with the choosen overloads
# and do any conversions that are needed.
@dataclass
class Replacer(visitor.Rewriter):
    net: PropagationNetwork = dataclasses.field(init=True)
    new_relations:OrderedSet[ir.Relation] = dataclasses.field(default_factory=OrderedSet[ir.Relation])

    def handle_model(self, model: ir.Model, parent: None):
        model = super().handle_model(model, parent)
        return model.reconstruct(
            relations=model.relations | self.new_relations
        )

    def handle_field(self, node: ir.Field, parent: ir.Node):
        if node in self.net.resolved_types:
            return f.field(node.name, self.net.resolved_types[node], node.input)
        return node

    def handle_var(self, node: ir.Var, parent: ir.Node):
        if node in self.net.resolved_types:
            return f.var(node.name, self.net.resolved_types[node])
        return node

    def handle_lookup(self, node: ir.Lookup, parent: ir.Node):
        # Boxing lookups are present to correctly guide type inference.
        # They should not propagate further down the pipeline.
        # Rewrite "type_box(from, to)" to "eq(from, to)"
        if helpers.is_type_box(node):
            annos = node.relation.annotations
            return f.lookup(builtins.eq, [node.args[0], node.args[1]], annos=list(annos))

        node = super().handle_lookup(node, parent)
        lookup_args = node.args

        # we can resolve to multiple overloads, that we'd need to union
        # so we'll need to hoist all the vars being output to make sure
        # everything joins correctly
        hoists = helpers.get_outputs(node)
        branches:list[ir.Logical] = []

        if node.relation == builtins.eq:
            tasks = []

            # We need to handle eq specially because its arguments can be converted symmetrically
            (left, right) = lookup_args
            left_type = to_type(left)
            right_type = to_type(right)
            mismatch = False
            if not type_matches(left_type, right_type):
                left_base = to_base_primitive(left_type)
                right_base = to_base_primitive(right_type)
                if conversion_allowed(left, left_type, right_type):
                    name = f"v{to_name(right_type)}"
                    if isinstance(left, ir.Var):
                        name = left.name + "_" + to_name(right_type)
                    new_left = f.var(name, right_type)
                    conv = CONVERSIONS[(left_base, right_base)]
                    self.new_relations.add(conv)
                    tasks.append(f.lookup(
                        conv,
                        [left, new_left]
                    ))
                    final_args = [new_left, right]
                elif conversion_allowed(right, right_type, left_type):
                    name = f"v{to_name(left_type)}"
                    if isinstance(right, ir.Var):
                        name = right.name + "_" + to_name(left_type)
                    new_right = f.var(name, left_type)
                    conv = CONVERSIONS[(right_base, left_base)]
                    self.new_relations.add(conv)
                    tasks.append(f.lookup(
                        conv,
                        [right, new_right]
                    ))
                    final_args = [left, new_right]
                else:
                    self.net.mismatch(node, left_type, right_type)
                    final_args = [left, right]
                    mismatch = True
            else:
                final_args = [left, right]

            if not mismatch:
                min_cost = float('inf')
                arg_types = set(to_type(arg) for arg in final_args)
                resolved = []
                for o in node.relation.overloads:
                    total = 0
                    for arg, field in zip(final_args, o.fields):
                        arg_type = to_type(arg)
                        field_type = to_type(field)
                        total += conversion_cost(arg, arg_type, field_type, arg_types)
                    if total <= min_cost:
                        if total < min_cost:
                            resolved.clear()
                            min_cost = total
                        resolved.append(o)

                if len(resolved) == 1:
                    tasks.append(f.lookup(resolved[0], final_args))
                else:
                    # If we cannot resolve the overload, just leave it.
                    self.net.unresolved_overload(node)
                    tasks.append(f.lookup(node.relation, final_args))
            else:
                # If there's a mismatch, just leave the original relation.
                tasks.append(f.lookup(node.relation, final_args))

            branches.append(f.logical(tasks, hoists))
        else:
            if node.id in self.net.resolved_overloads:
                resolved = self.net.resolved_overloads[node.id]
            else:
                resolved = [node.relation]

            for overload in resolved:
                # we may do conversions, so we can end up with multiple tasks
                # in this branch and we need to track what the final args are
                tasks = []
                final_args = []
                for arg, field in zip(lookup_args, overload.fields):
                    arg_type = to_type(arg)
                    field_type = to_type(field)
                    # the typer previously made sure that this should be valid so
                    # a type mismatch means we need to convert as long as this isn't
                    # a type variable
                    if field_type != types.EntityTypeVar and not type_matches(arg_type, field_type):
                        arg_base = to_base_primitive(arg_type)
                        field_base = to_base_primitive(field_type)
                        if (arg_base, field_base) in CONVERSIONS:
                            name = f"v{to_name(field_type)}"
                            if isinstance(arg, ir.Var):
                                name = arg.name + "_" + to_name(field_type)
                            new_arg = f.var(name, field.type)
                            conv = CONVERSIONS[(arg_base, field_base)]
                            self.new_relations.add(conv)
                            tasks.append(f.lookup(
                                conv,
                                [arg, new_arg]
                            ))
                            final_args.append(new_arg)
                        else:
                            final_args.append(arg)
                    else:
                        final_args.append(arg)
                tasks.append(f.lookup(overload, final_args))
                branches.append(f.logical(tasks, hoists))

        # unwrap if we don't actually need a union
        if len(branches) == 1:
            if len(branches[0].body) == 1:
                return branches[0].body[0]
            else:
                return branches[0]
        else:
            return f.union(branches, hoists)

#--------------------------------------------------
# Typer pass
#--------------------------------------------------

class InferTypes(compiler.Pass):
    def __init__(self):
        super().__init__()
        self.historical = {}

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        w = Analyzer()
        w.net.load_types(self.historical)

        # build the propagation network
        with debugging.span("type.analyze"):
            model.root.accept(w, model)

        # propagate the types through the network
        with debugging.span("type.propagate") as end_span:
            field_types = w.net.propagate()
            self.historical.update(field_types)
            end_span["type_graph"] = w.net.to_mermaid()
            end_span["type_report"] = w.net.to_fish()

        # replace the fields in the model with the new types
        with debugging.span("type.replace"):
            final = Replacer(w.net).walk(model)

        for err in w.net.errors:
            rich.print(str(err))

        return final
